"""
Test that the precessing SEOBNRv5PHM model has not changed.
Follows the procedure of IMRPhenomXPHM/IMRPhenomTPHM tests
in LALSuite.

This file was automatically generated on 2023-03-23 20:31:34.581748.
It should be regenerated using regenerate_SEOBNRv5PHM.py whenever
understood changes to the model are made.
"""

import pytest
import numpy as np
from pyseobnr.generate_waveform import generate_modes_opt, GenerateWaveform
import lal


def get_amp_phase(h):
    amp = np.abs(h)
    phase = np.unwrap(np.angle(h))
    return amp, phase


def sum_sqr_diff(x, y):
    return np.sqrt(np.sum((x - y) ** 2))


def gen_test_data(test_type):
    """
    compute the difference between two waveforms
    and compare to expected value
    """
    m1 = 50.0
    m2 = 30.0
    Mt = m1 + m2
    dt = 1 / 2048.0
    distance = 1.0
    inclination = np.pi / 3.0
    phiRef = 0.0
    approximant = "SEOBNRv5PHM"
    s1x = 0.5
    s1y = 0.3
    s2x = 0.3
    s2y = 0.7
    s1z = 0.2
    s2z = -0.1
    f_max = 1024.0
    f_min = 0.0157 / (Mt * np.pi * lal.MTSUN_SI)
    deltaF = 0.125
    params_dict = {
        "mass1": m1,
        "mass2": m2,
        "spin1x": s1x,
        "spin1y": s1y,
        "spin1z": s1z,
        "spin2x": s2x,
        "spin2y": s2y,
        "spin2z": s2z,
        "deltaT": dt,
        "deltaF": deltaF,
        "f22_start": f_min,
        "phi_ref": phiRef,
        "distance": distance,
        "inclination": inclination,
        "f_max": f_max,
        "approximant": approximant,
        "postadiabatic": True,
    }
    wfm_gen = GenerateWaveform(params_dict)  # We call the generator with the parameters
    params_dict2 = params_dict.copy()

    if test_type == "FD":
        # Test differences in Fourier domain, when varying intrinsic params
        hp1, hc1 = wfm_gen.generate_fd_polarizations()
        params_dict2.update({"mass2": 20.0})
        wfm_gen2 = GenerateWaveform(params_dict2)
        hp2, hc2 = wfm_gen2.generate_fd_polarizations()

        # Since we are in FD, hp,hc are *complex*, decompose into amp/phase
        hp1_amp, hp1_phase = get_amp_phase(hp1.data.data)
        hc1_amp, hc1_phase = get_amp_phase(hc1.data.data)

        hp2_amp, hp2_phase = get_amp_phase(hp2.data.data)
        hc2_amp, hc2_phase = get_amp_phase(hc2.data.data)

        hp_amp_diff = sum_sqr_diff(hp1_amp, hp2_amp)
        hp_phase_diff = sum_sqr_diff(hp1_phase, hp2_phase)

        hc_amp_diff = sum_sqr_diff(hc1_amp, hc2_amp)
        hc_phase_diff = sum_sqr_diff(hc1_phase, hc2_phase)

        return hp_amp_diff, hp_phase_diff, hc_amp_diff, hc_phase_diff

    elif test_type == "TD":
        # Test differences in time domain, when varying extrinsic params
        hp1, hc1 = wfm_gen.generate_td_polarizations()
        params_dict2.update({"inclination": 0.17})
        params_dict2.update({"phi_ref": 0.5})

        wfm_gen2 = GenerateWaveform(params_dict2)
        hp2, hc2 = wfm_gen2.generate_td_polarizations()
        # We are in TD, so hp,hc are *real*, compute differences directly
        hp_diff = sum_sqr_diff(hp1.data.data, hp2.data.data)
        hc_diff = sum_sqr_diff(hc1.data.data, hc2.data.data)
        return hp_diff, hc_diff


# -- test functions ---------------------


def test_SEOBNRv5PHM_diff_TD():
    """
    This test checks that SEOBNRv5HM hasn't changed in the time domain.
    It does this by generating two SEOBNRv5HM waveforms and computing
    their difference (according to their amplitude and phases)
    and compares them to pre-computed values.

    these pre-computed values were computed using the following line:

    `expected_result  =  np.array(gen_test_data("TD"))`

    """

    expected_result = np.array([9.286455953988354e-18, 8.842800671532437e-18])

    new_result = np.array(gen_test_data("TD"))
    np.testing.assert_allclose(
        new_result, expected_result, rtol=1e-4, err_msg="SEOBNRv5PHM TD test failed"
    )


def test_SEOBNRv5PHM_diff_FD():
    """
    This test checks that SEOBNRv5HM hasn't changed in frequency domain.
    It does this by generating two SEOBNRv5HM waveforms and computing
    their difference (according to their amplitude and phases)
    and compares them to pre-computed values.

    these pre-computed values were computed using the following line:

    `expected_result  =  np.array(gen_test_data("FD"))`

    """

    expected_result = np.array(
        [
            9.652274107044372e-20,
            3.391378195301547e04,
            8.605764565737375e-20,
            3.440922572212310e04,
        ]
    )

    new_result = np.array(gen_test_data("FD"))
    np.testing.assert_allclose(
        new_result, expected_result, rtol=4.5e-4, err_msg="SEOBNRv5PHM FD test failed"
    )



# cython: language_level=3

from pyseobnr.eob.hamiltonian.Hamiltonian_C cimport Hamiltonian_C

cpdef fin_diff_derivative(
    double[:] x,
    double[:] y,
    int n=?,
)

cpdef cumulative_integral(
    double[:] x,
    double[:] y,
    int order=?,
)

cpdef Kerr_ISCO(
    double chi1,
    double chi2,
    double m1,
    double m2,
)

cpdef compute_adiabatic_solution(
    double[:] r,
    Hamiltonian_C H,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    double tol=?,
)


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

# -- Project information -----------------------------------------------------

project = "pyseobnr"
copyright = "2022, pyseobnr authors"
author = "pyseobnr authors"

# The full version, including alpha/beta/rc tags
release = "0.0.1"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",  # needs to be first
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosectionlabel",
    "sphinx_tabs.tabs",
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx_gallery.load_style",
]

nbsphinx_custom_formats = {
    ".md": ["jupytext.reads", {"fmt": "mystnb"}],
}

autosectionlabel_prefix_document = True


autosummary_generate = True
# Add any paths that contain templates here, relative to this directory.
templates_path = ["templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "*.ipynb_checkpoints",
    "source/notebooks/*.ipynb",
    "source/notebooks/*.py",
    "README.md",
]
autoclass_content = "both"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_logo = "_static/pyseobnr_120px.png"



import argparse
import os
from pathlib import Path

from jinja2 import Environment, FileSystemLoader

this_folder = os.path.abspath(os.path.dirname(__file__))
script_home = os.path.join(this_folder, "../../")


if __name__ == "__main__":
    p = argparse.ArgumentParser()
    p.add_argument(
        "--n-cores",
        type=int,
        help="Number of cores to use. Should match number of tasks per node",
        default=64,
    )
    p.add_argument("--queue", type=str, help="The queue to use", default="syntrofos")
    p.add_argument(
        "--run-time",
        type=int,
        help="Number of hours to ask from the cluster",
        default=23,
    )
    p.add_argument(
        "--env", type=str, help="The path to the python env one wishes to use"
    )
    p.add_argument(
        "--test-dir",
        type=str,
        help="Name of directory where tests will be created",
        default="./",
    )
    p.add_argument("--points", type=int, help="Number of points", default=100000)
    p.add_argument(
        "--wrapper-path", type=str, help="The path to the wrapper, including name"
    )
    p.add_argument("--seed", type=int, help="Seed for random generation use", default=150914)
    args = p.parse_args()
    # Create submit files for the different tests
    file_loader = FileSystemLoader(f"{this_folder}/templates/")
    env = Environment(loader=file_loader)
    template = env.get_template("slurm.jinja")

    Path(args.test_dir).mkdir(parents=True, exist_ok=True)
    os.chdir(args.test_dir)

    # Amplitude hierarchy
    cmd = f"""python {script_home}auxiliary/sanity_checks/amplitude_hierarchy_test_precessing.py --points {args.points} --q-max 100 --a-max 0.998 --seed {args.seed} --wrapper-path {args.wrapper_path} """
    prog = template.render(
        job_name="amplitude_hierarchy_test_prec",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("amplitude_hierarchy.sh", "w") as fw:
        fw.write(prog)



import argparse
import importlib
import os
import sys

import numpy as np
from bilby.core.utils import check_directory_exists_and_if_not_mkdir
from pathos.multiprocessing import ProcessingPool as Pool

from pyseobnr.auxiliary.external_models import *
from pyseobnr.auxiliary.sanity_checks.metrics import *
from pyseobnr.auxiliary.sanity_checks.parameters import parameters_random_fast
from pyseobnr.generate_waveform import generate_modes_opt

modes_v5HM = [(2, 2), (2, 1), (3, 3), (3, 2), (4, 4), (4, 3), (5, 5)]
modes_v4HM = [(2, 2), (2, 1), (3, 3), (4, 4), (5, 5)]

# Reproducible
seed = 150914

def mismatch_EOB(q, chi1, chi2, model_1_name, model_2_name):

    omega0 = 0.015

    if model_1_name == "SEOBNRv5HM":
        _,_,model_1 = generate_modes_opt(q,chi1,chi2,0.8*omega0,debug=True)

    elif model_1_name == "SEOBNRv4HM":
        model_1 = SEOBNRv4HM_LAL(q, chi1, chi2, 0.8*omega0)
        model_1()


    model_2 = model_2_name

    masses = np.arange(10, 310, 10)

    unf_settings = {"debug": True, "masses": masses}
    unf = UnfaithfulnessModeByModeLAL(settings=unf_settings)

    modes = [(2,2)]
    mms = []
    params = dict(
                q=q,
                chi1=np.array([0.0, 0.0, chi1]),
                chi2=np.array([0.0, 0.0, chi2]),
                omega0= omega0,
                dt=model_1.delta_T,
                df=1 / (len(model_1.t) * model_1.delta_T),
    )

    for mode in modes:
        ell, m = mode
        mm = unf(model_1, model_2, params=params, ell=ell, m=m)
        mms.append(mm)

    return (q, chi1, chi2, np.array(mms))


def process_one_case(input):
    q, chi1, chi2, model_1_name, model_2_name = input
    q, chi1, chi2, mm = mismatch_EOB(q, chi1, chi2, model_1_name, model_2_name)
    return mm


if __name__ == "__main__":
    p = argparse.ArgumentParser(description="Compute mismatch between EOB and Phenom models")
    p.add_argument("--points", type=int, help="Number of points", default="5000")
    p.add_argument("--chi-max", type=float, help="Maximum spin", default="0.995")
    p.add_argument("--q-min", type=float, help="Minimum mass-ratio", default="1.0")
    p.add_argument("--q-max", type=float, help="Maximum mass-ratio", default="100.0")
    p.add_argument(
        "--name",
        type=str,
        help="Name of the output file",
        default="statistics_EOB_comparable",
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    p.add_argument(
        "--model-1-name",
        type=str,
        help="Name of the first EOB model",
        default="SEOBNRv5HM",
    )
    p.add_argument(
        "--model-2-name",
        type=str,
        help="Name of the second Phenom model. Must be one of 'IMRPhenomT', 'IMRPhenomXAS'",
        default="IMRPhenomXAS",
    )
    p.add_argument("--n-cpu", type=int, help="Number of cores to use", default=64)
    p.add_argument(
        "--plots-only",
        action="store_true",
        help="Only generate plots, don't recompute things",
    )
    p.add_argument(
        "--include-all",
        action="store_true",
        help="Include odd m modes close to equal mass",
        default=False,
    )

    args = p.parse_args()

    qarr, chi1arr, chi2arr = parameters_random_fast(
        args.points,
        args.q_min,
        args.q_max,
        -args.chi_max,
        args.chi_max,
        -args.chi_max,
        args.chi_max,
        random_state=seed
    )
    lst = [
        (q, chi1, chi2, args.model_1_name, args.model_2_name)
        for q, chi1, chi2 in zip(qarr, chi1arr, chi2arr)
    ]

    pool = Pool(args.n_cpu)
    all_means = list(pool.map(process_one_case, lst))

    mode_list = [(2,2)]

    aall_means = np.array(all_means)
    # This is an array with shape n cases x p modes x m masses
    mismatches = {}
    for mode in mode_list:
        mismatches[mode] = []
    for i in range(len(all_means)):
        mm_for_case = all_means[i]
        for j, mode in enumerate(mode_list):
            mode_mismatch = mm_for_case[j]
            mismatches[mode].append(mode_mismatch)

    for mode in mode_list:
        ell, m = mode
        np.savetxt(
            f"{args.name}_{args.model_1_name}_{args.model_2_name}{ell}{m}.dat",
            mismatches[mode],
        )

    np.savetxt(
        f"parameters_{args.model_1_name}_{args.model_2_name}_{args.name}.dat",
        np.c_[qarr, chi1arr, chi2arr],
    )
    ### Plots ###

    if args.plots:

        import matplotlib
        import matplotlib.pyplot as plt

        plt_dir = "./plots"
        check_directory_exists_and_if_not_mkdir(plt_dir)

        res_path = args.name
        params = np.genfromtxt(
            f"parameters_{args.model_1_name}_{args.model_2_name}_{args.name}.dat"
        )
        q = params[:, 0]
        chi1 = params[:, 1]
        chi2 = params[:, 2]

        nu = q / (1 + q) ** 2
        m1 = q / (1 + q)
        m2 = 1 / (1 + q)
        ap = m1 * chi1 + m2 * chi2
        am = m1 * chi1 - m2 * chi2
        for mode in mode_list:
            ell, m = mode
            mm_M = np.loadtxt(
                f"{args.name}_{args.model_1_name}_{args.model_2_name}{ell}{m}.dat"
            )
            if not args.include_all:
                if m % 2 == 1:
                    idx_notEMES = np.where((q > 1.01) | (np.abs(chi1 - chi2) > 0.01))[0]
                    q_pl = q[idx_notEMES]
                    ap_pl = ap[idx_notEMES]
                    mm_M = mm_M[idx_notEMES]
                else:
                    q_pl = q
                    ap_pl = ap
            else:
                q_pl = q
                ap_pl = ap

            M = np.arange(10, 310, 10)

            # maximum of mm_M across total mass for histogram
            mm = []
            for mp in mm_M:
                mm.append(np.max(mp))
            mm = np.array(mm)

            # Spaghetti plot
            for mp in mm_M:
                plt.plot(M, mp, color="C0", linewidth=0.5)
            plt.axhline(0.01, ls="--", color="red")
            plt.yscale("log")
            plt.xlabel("M")
            plt.ylabel("$\mathcal{M}$")
            plt.xlim(10, 300)
            plt.savefig(
                f"{plt_dir}/mm_spaghetti_{args.model_1_name}_{args.model_2_name}_{args.name}{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

        # Histogram
        plt.hist(
            mm,
            bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=50),
            alpha=0.4,
            label=args.model_1_name + " - " + args.model_2_name,
        )
        plt.axvline(np.median(mm), c="C0", ls="--")
        plt.legend(loc="best")
        plt.gca().set_xscale("log")
        plt.xlabel("$\mathcal{M}_{\mathrm{Max}}$")
        plt.title(
            "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),6)}"
        )
        plt.savefig(
            f"{plt_dir}/mm_hist_{args.model_1_name}_{args.model_2_name}_{args.name}{ell}{m}.png",
            bbox_inches="tight",
            dpi=300,
        )
        plt.close()

        # CDF
        plt.hist(
            mm,
            bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=100),
            alpha=0.4,
            label=args.model_1_name + " - " + args.model_2_name,
            cumulative=True,
            density=True,
            histtype="step",
            lw=2,
        )
        plt.legend(loc="best")
        plt.gca().set_xscale("log")
        plt.xlabel("$\mathcal{M}_{\mathrm{Max}}$")
        plt.title(
            "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),6)}"
        )
        plt.grid(True, which="both", ls=":")
        plt.savefig(
            f"{plt_dir}/mm_cdf_{args.model_1_name}_{args.model_2_name}_{args.name}{ell}{m}.png",
            bbox_inches="tight",
            dpi=300,
        )

        plt.close()

        # Mismatch across parameter space
        mm_s, q_s, ap_s = map(list, zip(*sorted(zip(mm, q_pl, ap_pl))))
        plt.scatter(
            q_s, ap_s, c=mm_s, linewidths=1, norm=matplotlib.colors.LogNorm()
        )
        plt.ylabel("$\chi_{\mathrm{eff}}$")
        plt.xlabel("$q$")
        cbar = plt.colorbar()
        cbar.set_label("$\mathcal{M}$")
        plt.savefig(
            f"{plt_dir}/mm_scatter_{args.model_1_name}_{args.model_2_name}_{args.name}{ell}{m}.png",
            bbox_inches="tight",
            dpi=300,
        )
        plt.close()



"""
Mode mixing fitting function and coefficients from [Berti2014]_.

Used to model mode mixing in the (3,2) and (4,3) modes.
"""

import numpy as np
from numba import jit


@jit(nopython=True)
def mu_fit(j: float, p1: float, p2: float, p3: float, p4: float) -> float:
    """
    Equation 11 from [Berti2014]_ to get the spherical-spheroidal mode mixing coefficients

    Args:
        j (float): dimensionless spin parameter
        p1 (float): first fitting parameter
        p2 (float): second fitting parameter
        p3 (float): third fitting parameter
        p4 (float): fourth fitting parameter

    Returns:
        float: the spherical-spheroidal mode mixing coefficient
    """
    return p1 * abs(j) ** (p2) + p3 * abs(j) ** (p4)

@jit(nopython=True)
def mu(m: int, l: int, lp: int, j: float) -> float:
    """
    Equation 11 from [Berti2014]_ applied to the (2,2), (3,2), (3,3) and (4,3) modes using the
    appropriate fitting parameters

    Data files containing the fits https://git.ligo.org/waveforms/reviews/seobnrv5/-/blob/main/aligned/docs/swsh_fits.dat
    (copied from https://pages.jh.edu/eberti2/ringdown/ on 02/12/2022)

    Args:
        m (int): m index of the relevant mode
        l (int): :math:`\\ell` index of the relevant mode
        lp (int): index of modes with the same m and lp<=l which give rise to mode mixing
        j (float): dimensionless spin parameter

    Returns:
        float: the spherical-spheroidal mode mixing coefficient for the specified mode
    """

    # (3,2) mode

    if m == 2 and l == 2 and lp == 2:
        return (
            1
            + mu_fit(j, -7.39789e-03, 2.88933e00, -6.61369e-03, 1.71287e01)
            + 1j * mu_fit(j, 1.53046e-02, 1.21928e00, -9.34293e-03, 2.49915e01)
        )

    elif m == 2 and l == 3 and lp == 2:
        return (
            mu_fit(j, -1.03512e-01, 1.22285e00, -5.74989e-02, 8.70536e00)
            + 1j * mu_fit(j, -1.60040e-02, 9.53385e-01, 1.00344e-02, 1.47550e01)
        ) * -1
        # see footnote 4 of 1902.02731
    elif m == 2 and l == 3 and lp == 3:
        return (
            1
            + mu_fit(j, -2.51626e-02, 2.43323e00, -1.30029e-02, 1.09648e01)
            + 1j * mu_fit(j, -3.23812e-02, 9.21248e-01, 1.88165e-02, 1.06988e01)
        )

    # (4,3) mode

    elif m == 3 and l == 3 and lp == 3:
        return (
            1
            + mu_fit(j, -1.20858e-02, 2.94266e00, -1.12402e-02, 1.72866e01)
            + 1j * mu_fit(j, -1.72988e-02, 8.51080e-01, 1.43925e-02, 5.98677e00)
        )

    elif m == 3 and l == 4 and lp == 3:
        return (
            mu_fit(j, -1.32724e-01, 1.24145e00, -7.34932e-02, 8.51670e00)
            + 1j * mu_fit(j, -1.15656e-02, 9.01101e-01, 7.32291e-03, 1.17007e01)
        ) * -1
        # see footnote 4 of 1902.02731
    elif m == 3 and l == 4 and lp == 4:
        return (
            1
            + mu_fit(j, -3.42425e-02, 2.59986e00, -2.12509e-02, 1.25368e01)
            + 1j * mu_fit(j, -4.54823e-02, 9.25823e-01, 2.84355e-02, 1.14204e01)
        )

    else:
        raise NotImplementedError("Coefficient not implemented.")


# Auxiliary functions to obtain the spheroidal mode input values
# See Eq. 73 to Eq. 80 of [SEOBNRv5HM-notes]_


def rho(ell, m, j, h_ellm, h_mm):
    """
    Auxiliary function to obtain the spheroidal mode input values, see Eq.(73) of [SEOBNRv5HM-notes]_

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        h_ellm (float): amplitude of the (:math:`\\ell,m`) mode at attachment time
        h_mm (float): amplitude of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the value of rho
    """
    return abs(mu(m, ell, m, j)) * (h_mm) / (h_ellm * abs(mu(m, m, m, j)))


def dphi(ell, m, j, phi_ellm, phi_mm):
    """
    Auxiliary function to obtain the spheroidal mode input values, see Eq.(74) of [SEOBNRv5HM-notes]_

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        phi_ellm (float): phase of the (:math:`\\ell,m`) mode at attachment time
        phi_mm (float): phase of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the value of dphi
    """
    return phi_mm - phi_ellm - np.angle(mu(m, ell, m, j)) + np.angle(mu(m, m, m, j))


def F(ell, m, j, h_ellm, h_mm, phi_ellm, phi_mm):
    """
    Auxiliary function to obtain the spheroidal mode input values, see Eq.(75) of [SEOBNRv5HM-notes]_

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        h_ellm (float): amplitude of the (:math:`\\ell,m`) mode at attachment time
        h_mm (float): amplitude of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        phi_ellm (float): phase of the (:math:`\\ell,m`) mode at attachment time
        phi_mm (float): phase of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the value of F
    """
    return np.sqrt(
        (1 - rho(ell, m, j, h_ellm, h_mm) * np.cos(dphi(ell, m, j, phi_ellm, phi_mm)))
        ** 2
        + (rho(ell, m, j, h_ellm, h_mm) ** 2)
        * (np.sin(dphi(ell, m, j, phi_ellm, phi_mm))) ** 2
    )


def alpha(ell, m, j, h_ellm, h_mm, phi_ellm, phi_mm):
    """
    Auxiliary function to obtain the spheroidal mode input values, see Eq.(76) of [SEOBNRv5HM-notes]_

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        h_ellm (float): amplitude of the (:math:`\\ell,m`) mode at attachment time
        h_mm (float): amplitude of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        phi_ellm (float): phase of the (:math:`\\ell,m`) mode at attachment time
        phi_mm (float): phase of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the value of alpha
    """
    return np.arctan2(
        -rho(ell, m, j, h_ellm, h_mm) * np.sin(dphi(ell, m, j, phi_ellm, phi_mm)),
        (1 - rho(ell, m, j, h_ellm, h_mm) * np.cos(dphi(ell, m, j, phi_ellm, phi_mm))),
    )

@jit(nopython=True)
def rho_dot(ell, m, j, h_ellm, h_mm, hdot_ellm, hdot_mm):
    """
    Auxiliary function to obtain the spheroidal mode input values, see Eq.(77) of [SEOBNRv5HM-notes]_

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        h_ellm (float): amplitude of the (:math:`\\ell,m`) mode at attachment time
        h_mm (float): amplitude of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        hdot_ellm (float): amplitude's first derivative of the (:math:`\\ell,m`) mode at attachment time
        hdot_mm (float): amplitude's first derivative of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the value of rho_dot
    """
    return abs(mu(m, ell, m, j)) * (hdot_mm / h_ellm - h_mm * hdot_ellm / (h_ellm ** 2))

@jit(nopython=True)
def dphi_dot(omega_ellm, omega_mm):
    """
    Auxiliary function to obtain the spheroidal mode input values, see Eq.(78) of [SEOBNRv5HM-notes]_

    Args:
        omega_ellm (float): frequency of the (:math:`\\ell,m`) mode at attachment time
        omega_mm (float): frequency of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the value of dphi_dot
    """
    return omega_mm - omega_ellm

def F_dot(
    ell, m, j, h_ellm, h_mm, hdot_ellm, hdot_mm, omega_ellm, omega_mm, phi_ellm, phi_mm
):
    """
    Auxiliary function to obtain the spheroidal mode input values, see Eq.(79) of [SEOBNRv5HM-notes]_

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        h_ellm (float): amplitude of the (:math:`\\ell,m`) mode at attachment time
        h_mm (float): amplitude of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        hdot_ellm (float): amplitude's first derivative of the (:math:`\\ell,m`) mode at attachment time
        hdot_mm (float): amplitude's first derivative of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        omega_ellm (float): frequency of the (:math:`\\ell,m`) mode at attachment time
        omega_mm (float): frequency of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        phi_ellm (float): phase of the (:math:`\\ell,m`) mode at attachment time
        phi_mm (float): phase of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the value of F_dot
    """
    return (
        rho(ell, m, j, h_ellm, h_mm)
        * rho_dot(ell, m, j, h_ellm, h_mm, hdot_ellm, hdot_mm)
        + rho(ell, m, j, h_ellm, h_mm)
        * np.sin(dphi(ell, m, j, phi_ellm, phi_mm))
        * dphi_dot(omega_ellm, omega_mm)
        - rho_dot(ell, m, j, h_ellm, h_mm, hdot_ellm, hdot_mm)
        * np.cos(dphi(ell, m, j, phi_ellm, phi_mm))
    ) / F(ell, m, j, h_ellm, h_mm, phi_ellm, phi_mm)

def alpha_dot(
    ell, m, j, h_ellm, h_mm, hdot_ellm, hdot_mm, omega_ellm, omega_mm, phi_ellm, phi_mm
):
    """
    Auxiliary function to obtain the spheroidal mode input values, see Eq.(80) of [SEOBNRv5HM-notes]_

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        h_ellm (float): amplitude of the (:math:`\\ell,m`) mode at attachment time
        h_mm (float): amplitude of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        hdot_ellm (float): amplitude's first derivative of the (:math:`\\ell,m`) mode at attachment time
        hdot_mm (float): amplitude's first derivative of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        omega_ellm (float): frequency of the (:math:`\\ell,m`) mode at attachment time
        omega_mm (float): frequency of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        phi_ellm (float): phase of the (:math:`\\ell,m`) mode at attachment time
        phi_mm (float): phase of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the value of alpha_dot
    """
    return (
        rho(ell, m, j, h_ellm, h_mm) ** 2 * dphi_dot(omega_ellm, omega_mm)
        - rho(ell, m, j, h_ellm, h_mm)
        * np.cos(dphi(ell, m, j, phi_ellm, phi_mm))
        * dphi_dot(omega_ellm, omega_mm)
        - rho_dot(ell, m, j, h_ellm, h_mm, hdot_ellm, hdot_mm)
        * np.sin(dphi(ell, m, j, phi_ellm, phi_mm))
    ) / (F(ell, m, j, h_ellm, h_mm, phi_ellm, phi_mm) ** 2)


# Spheroidal mode input values
# See Eq. 81 to Eq. 84 of [SEOBNRv5HM-notes]_

def h_ellm0_nu(ell, m, j, h_ellm, h_mm, phi_ellm, phi_mm):
    """
    Computes the amplitude of the spheroidal :math:`(\\ell,m,0)` mode at at attachment time
    from the :math:`(\\ell,m)` and (:math:`\\ell^{\\prime} = m,m`) spherical modes

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        h_ellm (float): amplitude of the (:math:`\\ell,m`) mode at attachment time
        h_mm (float): amplitude of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        phi_ellm (float): phase of the (:math:`\\ell,m`) mode at attachment time
        phi_mm (float): phase of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the amplitude of the spheroidal :math:`(\\ell,m,0)` mode at at attachment time
    """
    return (
        h_ellm
        * F(ell, m, j, h_ellm, h_mm, phi_ellm, phi_mm)
        / np.abs(mu(m, ell, ell, j))
    )

def phi_ellm0(ell, m, j, h_ellm, h_mm, phi_ellm, phi_mm):
    """
    Computes the phase of the spheroidal :math:`(\\ell,m,0)` mode at at attachment time
    from the (:math:`\\ell,m`) and (:math:`\\ell^{\\prime} = m,m`) spherical modes

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        h_ellm (float): amplitude of the (:math:`\\ell,m`) mode at attachment time
        h_mm (float): amplitude of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        phi_ellm (float): phase of the (:math:`\\ell,m`) mode at attachment time
        phi_mm (float): phase of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the phase of the spheroidal :math:`(\\ell,m,0)` mode at at attachment time
    """
    return (
        phi_ellm
        + np.angle(mu(m, ell, ell, j))
        + alpha(ell, m, j, h_ellm, h_mm, phi_ellm, phi_mm)
    )

def hdot_ellm0_nu(
    ell, m, j, h_ellm, h_mm, hdot_ellm, hdot_mm, omega_ellm, omega_mm, phi_ellm, phi_mm
):  
    """
    Computes the amplitude's first derivative of the spheroidal :math:`(\\ell,m,0)` mode at at attachment time
    from the (:math:`\\ell,m`) and (:math:`\\ell^{\\prime} = m,m`) spherical modes

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        h_ellm (float): amplitude of the (:math:`\\ell,m`) mode at attachment time
        h_mm (float): amplitude of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        hdot_ellm (float): amplitude's first derivative of the (:math:`\\ell,m`) mode at attachment time
        hdot_mm (float): amplitude's first derivative of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        omega_ellm (float): frequency of the (:math:`\\ell,m`) mode at attachment time
        omega_mm (float): frequency of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        phi_ellm (float): phase of the (:math:`\\ell,m`) mode at attachment time
        phi_mm (float): phase of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the amplitude's first derivative of the spheroidal :math:`(\\ell,m,0)` mode at at attachment time
    """
    return (
        hdot_ellm * F(ell, m, j, h_ellm, h_mm, phi_ellm, phi_mm)
        + h_ellm
        * F_dot(
            ell,
            m,
            j,
            h_ellm,
            h_mm,
            hdot_ellm,
            hdot_mm,
            omega_ellm,
            omega_mm,
            phi_ellm,
            phi_mm,
        )
    ) / np.abs(mu(m, ell, ell, j))

def omega_ellm0(
    ell, m, j, h_ellm, h_mm, hdot_ellm, hdot_mm, omega_ellm, omega_mm, phi_ellm, phi_mm
):
    """
    Computes the frequency of the spheroidal :math:`(\\ell,m,0)` mode at at attachment time
    from the (:math:`\\ell,m`) and (:math:`\\ell^{\\prime} = m,m`) spherical modes

    Args:
        ell (int): :math:`\\ell` index of the relevant mode
        m (int): m index of the relevant mode
        j (float): dimensionless spin parameter
        h_ellm (float): amplitude of the (:math:`\\ell,m`) mode at attachment time
        h_mm (float): amplitude of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        hdot_ellm (float): amplitude's first derivative of the (:math:`\\ell,m`) mode at attachment time
        hdot_mm (float): amplitude's first derivative of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        omega_ellm (float): frequency of the (:math:`\\ell,m`) mode at attachment time
        omega_mm (float): frequency of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time
        phi_ellm (float): phase of the (:math:`\\ell,m`) mode at attachment time
        phi_mm (float): phase of the (:math:`\\ell^{\\prime} = m,m`) mode at attachment time

    Returns:
        float: the frequency of the spheroidal :math:`(\\ell,m,0)` mode at at attachment time
    """
    return omega_ellm + alpha_dot(
        ell,
        m,
        j,
        h_ellm,
        h_mm,
        hdot_ellm,
        hdot_mm,
        omega_ellm,
        omega_mm,
        phi_ellm,
        phi_mm,
    )



"""
Computes the quasi-precessing initial conditions in polar coordinates.

These have the same equations as the aligned-spin, however:
- the Hamiltonian is different because it depends on several spin degrees of freedom
- appropriate spin projections (and spins themselves) are passed to the Hamiltonian and the RR forces
"""

from typing import Callable

import numpy as np
from scipy.optimize import root, root_scalar

from ..hamiltonian import Hamiltonian


def IC_cons_augm(
    u: np.ndarray,
    omega: float,
    H: Hamiltonian,
    chi1_v: np.ndarray,
    chi2_v: np.ndarray,
    chi1_LN: float,
    chi2_LN: float,
    chi1_L: float,
    chi2_L: float,
    m_1: float,
    m_2: float,
):
    """The equations defining the 'conservative'
    part of the QC initial conditions, namely
    for r and pphi.

    These are Eq(4.8,4.9) in [Buonanno2005]_.

    Args:
        u ([np.ndarray]): The unknowns, r,pphi
        omega (float): Desired starting orbital frequency, in geometric units
        H (function): The Hamiltonian to use (an instance of Hamiltonian class)
        chi1_v (np.ndarray): Dimensionless spin vector of the primary
        chi2_v (np.ndarray): Dimensionless spin vector of the secondary
        chi1_LN (float): Projection of the dimensionless spin vector of the primary onto LN
        chi2_LN (float): Projection of the dimensionless spin vector of the secondary onto LN
        chi1_L (float): Projection of the dimensionless spin vector of the primary onto L
        chi2_L (float): Projection of the dimensionless spin vector of the secondary onto L
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary

    Returns:
        [np.ndarray]: The desired equations evaluated at u
    """
    r, pphi = u

    q = np.array([r, 0.0])
    p = np.array([0.0, pphi])

    grad = H.dynamics(q, p, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L)
    dHdr = grad[0]
    dHdpphi = grad[3]
    diff = np.array([dHdpphi - omega, dHdr])
    return diff


def IC_diss_augm(
    u: np.ndarray,
    r: float,
    pphi: float,
    H: Hamiltonian,
    RR: Callable,
    chi1_v: np.ndarray,
    chi2_v: np.ndarray,
    chi1_LN: float,
    chi2_LN: float,
    chi1_L: float,
    chi2_L: float,
    m_1: float,
    m_2: float,
    params,
):
    """Initial conditions for the "dissipative" part,
    namely pr

    This is basically Eq(4.15) in [Buonanno2005]_.

    Args:
        u (float): Guess for pr
        r (float): Starting separation
        pphi (float): Starting angular momentum
        H (Hamiltonian): The Hamiltonian object to use
        RR (Callable): Function that returns the RR force. Must have same signature as the Hamiltonian
        chi1_v (np.ndarray): Dimensionless spin vector of the primary
        chi2_v (np.ndarray): Dimensionless spin vector of the secondary
        chi1_LN (float): Projection of the dimensionless spin vector of the primary onto LN
        chi2_LN (float): Projection of the dimensionless spin vector of the secondary onto LN
        chi1_L (float): Projection of the dimensionless spin vector of the primary onto L
        chi2_L (float): Projection of the dimensionless spin vector of the secondary onto L
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary
        params (EOBParams): container with additional useful quantities

    Returns:
        float: The equation for pr
    """
    pr = u
    q = np.array([r, 0.0])
    p = np.array([pr, pphi])
    hess = H.hessian(q, p, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L)
    d2Hdr2 = hess[0, 0]
    d2HdrdL = hess[3, 0]
    dLdr = -d2Hdr2 / d2HdrdL
    p_circ = np.array([0.0, p[1]])
    dynamics = H.dynamics(
        q, p, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L
    )
    H_val = dynamics[4]
    omega = dynamics[3]
    omega_circ = H.omega(
        q, p_circ, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L
    )
    RR_f = RR.RR(q, p, omega, omega_circ, H_val, params)
    csi = dynamics[5]

    rdot = 1 / csi * RR_f[1] / dLdr
    dHdpr = dynamics[2]
    return rdot - dHdpr


def computeIC_augm(
    omega: float,
    H: Hamiltonian,
    RR: Callable,
    chi1_v: np.ndarray,
    chi2_v: np.ndarray,
    m_1: float,
    m_2: float,
    **kwargs,
):
    """Compute the initial conditions for an aligned-spin BBH binary

    Args:
        omega (float): Initial *orbital* frequency in geometric units
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR force to use
        chi1_v (np.ndarray): Dimensionless spin vector of the primary
        chi2_v (np.ndarray): Dimensionless spin vector of the secondary
        m_1 (float): Mass of primary
        m_2 (float): Mass of secondary

    Returns:
        tuple: The initial conditions: (r,pphi,pr)
    """

    params = kwargs["params"]

    chi1_LN, chi2_LN = params.p_params.chi_1, params.p_params.chi_2
    chi1_L, chi2_L = params.p_params.chi1_L, params.p_params.chi2_L

    # Initial guess from Newtonian gravity
    r_guess = omega ** (-2.0 / 3)
    z = [r_guess, np.sqrt(r_guess)]

    # The conservative bit: solve for r and pphi
    res_cons = root(
        IC_cons_augm,
        z,
        args=(omega, H, chi1_v, chi2_v, chi1_LN, chi2_LN, chi1_L, chi2_L, m_1, m_2),
        tol=1e-12,
    )
    r0, pphi0 = res_cons.x

    res_diss = root_scalar(
        IC_diss_augm,
        bracket=[-3e-2, 0],
        args=(
            r0,
            pphi0,
            H,
            RR,
            chi1_v,
            chi2_v,
            chi1_LN,
            chi2_LN,
            chi1_L,
            chi2_L,
            m_1,
            m_2,
            kwargs["params"],
        ),
        xtol=1e-12,
        rtol=1e-10,
    )

    # Now do the dissipative bit: solve for pr
    pr0 = res_diss.root

    # Evaluate omega_circ and the Hamiltonian at the intial conditions
    q = np.array([r0, 0.0])
    p = np.array([pr0, pphi0])
    p_circ = np.array([0.0, p[1]])
    dynamics = H.dynamics(
        q, p, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L
    )
    H_val = dynamics[4]
    omega_circ = H.omega(
        q, p_circ, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L
    )
    params.p_params.omega_circ = omega_circ
    params.p_params.H_val = H_val

    # print(f"Ad. IC : r0 = {r0}, pphi0 = {pphi0}, pr0 = {pr0}")

    return r0, pphi0, pr0



"""
Additional utility functions to manipulate various aspects of precession dynamics.
"""

from math import cos, sqrt
from typing import Any, Dict

import lal
import numpy as np
import quaternion
from numba import *
from numba import jit, types
from scipy.interpolate import CubicSpline
from scri import WaveformModes

from ..fits.EOB_fits import compute_QNM
from ..fits.fits_Hamiltonian import dSO as dSO_poly_fit
from ..hamiltonian import Hamiltonian
from .math_ops_opt import my_cross, my_dot, my_norm

#################################################################################################################

###########              FUNCTIONS TO  APPLY THE MERGER RINGDOWN APPROXIMATION FOR THE ANGLES          ##########

#################################################################################################################


# This function does exactly the same as the LAL counterpart
def SEOBBuildJframeVectors(Jhat_final: np.ndarray):
    """
    This function computes the Jframe unit vectors, with e3J along Jhat.
    Convention: if (ex, ey, ez) is the initial I-frame, e1J chosen such that ex
    is in the plane (e1J, e3J) and ex.e1J>0.
    In the case where e3J and x happen to be close to aligned, we continuously
    switch to another prescription with y playing the role of x.
    Same operation as in SEOBBuildJframeVectors in
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimIMRSpinPrecEOBv4P.c#L2833
    LALSimIMRSpinPrecEOBv4P.c.
    See Appendix A of https://arxiv.org/abs/1607.05661v1
    and Sec. III B of https://arxiv.org/pdf/2004.09442v1.pdf for details.

    Args:
        Jhat_final(np.ndarray): Final total angular momentum vector

    Returns:
        (tuple): Triad of unit vectors defining the J-frame
    """

    e3J = Jhat_final

    exvec = np.array([1.0, 0.0, 0.0])
    eyvec = np.array([0.0, 1.0, 0.0])

    exdote3J = my_dot(exvec, e3J)
    eydote3J = my_dot(eyvec, e3J)

    lambda_fac = 1.0 - abs(exdote3J)

    if lambda_fac < 0.0 or lambda_fac > 1.0:
        print("Problem: lambda=1-|e3J.ex|=%f, should be in [0,1]" % lambda_fac)

    elif lambda_fac > 1e-4:
        normfacx = 1.0 / sqrt(1.0 - exdote3J * exdote3J)

        e1J = (exvec - exdote3J * e3J) / normfacx
    elif lambda_fac < 1e-5:

        normfacy = 1.0 / sqrt(1.0 - eydote3J * eydote3J)
        e1J = (eyvec - eydote3J * e3J) / normfacy
    else:
        weightx = (lambda_fac - 1e-5) / (1e-4 - 1e-5)
        weighty = 1.0 - weightx
        normfacx = 1.0 / sqrt(1.0 - exdote3J * exdote3J)
        normfacy = 1.0 / sqrt(1.0 - eydote3J * eydote3J)
        e1J = (
            weightx * (exvec - exdote3J * e3J) / normfacx
            + weighty * (eyvec - eydote3J * e3J) / normfacy
        )

        e1Jblendednorm = my_norm(e1J)
        e1J /= e1Jblendednorm

    # Get e2J = e3J x e1J
    e2J = my_cross(e3J, e1J)

    e1Jnorm = my_norm(e1J)
    e2Jnorm = my_norm(e2J)
    e3Jnorm = my_norm(e3J)

    e1J /= e1Jnorm
    e2J /= e2Jnorm
    e3J /= e3Jnorm
    #print(f"e1J:{e1J},e2J:{e2J},e3J:{e3J}")
    return e1J, e2J, e3J


def SEOBEulerJ2PFromDynamics(
    t: np.ndarray,
    omega: np.ndarray,
    LNhat: np.ndarray,
):
    """
    This function computes the Euler angles from J-frame to P-frame given the
    Newtonian orbital angular momentum timeseries LNhat(t) in the J frame.
    Similar operation as SEOBEulerJ2PFromDynamics in LALSimIMRSpinPrecEOBv4P.c.
    See Appendix A of https://arxiv.org/abs/1607.05661v1
    and Sec. III B of https://arxiv.org/pdf/2004.09442v1.pdf for details.

    Args:
        t (np.ndarray): Time array
        omega (np.ndarray): EOB orbital frequency
        LNhat (np.ndarray): LNhat vector array array (in J frame!)

    Returns:
        (tuple): Time-dependent quaterions describing the rotation from the co-precessing frame to the J-frame

    """

    LN_in_J_quat = quaternion.as_quat_array(np.c_[np.zeros(len(LNhat)), LNhat])
    # Compute dot product between LNhat and Jf (remember we are in J frame!)
    dtprd = np.einsum("ij,j->i", LNhat, np.array([0, 0, 1.0]))
    # Check if at any point these vectors are aligned or anti-aligned
    # If this is the case we will need to be careful when constructing the
    # rotation from J frame to P frame to avoid numerical issues
    # If the vectors are almost colinear we split the rotation into 2
    # steps: a rotation of Jf into an auxiliary vector aux and a
    # rotation of aux into LN. This ensures that the rotation is smooth
    # without artefacts to due to roundoff
    if np.amin(np.abs(np.abs(dtprd) - 1)) < 1e-3:
        # Pick the x axis of the J frame as the auxiliary vector
        # This should be a good choice since it's perpendicular to
        # final spin
        aux = quaternion.quaternion(0.0, 1.0, 0.0, 0.0)
        # Still, check if LN is degenerate with x axis of the J frame
        dtprd2 = np.einsum("ij,j->i", LNhat, np.array([1, 0, 0.0]))
        if np.amin(np.abs(np.abs(dtprd2) - 1)) < 1e-3:
            # If not x, then y
            aux = quaternion.quaternion(0.0, 0.0, 1.0, 0.0)
    else:
        aux = None
    
    # Note that because we are working in the J frame the final spin
    # direction is just along z axis, by definition.
    quatJ2P = minimal_quat(quaternion.z, LN_in_J_quat, aux, t, omega)

    return quatJ2P


def SEOBRotatehIlmFromhJlm_opt_v1(
    w_hJlm: WaveformModes,
    modes_lmax: int,
    quatI2J: quaternion,
) -> WaveformModes:

    """
     This function computes the hIlm Re/Im timeseries (fixed sampling) from hJlm
     Re/Im timeseries (same sampling). This is a simple rotation,
     sample-by-sample, with constant Wigner coefficients.

     Args:
        w_hJlm (WaveformModes): WaveformModes object from the scri python package. It contains waveform modes in the final J-frame
        modes_lmax (int): Maximum value of l in modes (l,m)
        quatI2J (float): Quaternion between the I-frame and the J-frame

    Returns:
        (WaveformModes): WaveformModes scri object in the inertial frame

    """
    res = w_hJlm.rotate_decomposition_basis(~quatI2J)

    return res


def seobnrv4P_quaternionJ2P_postmerger_extension(
    t_full: np.ndarray,
    precRate: float,
    euler_angles_attach: np.ndarray,
    euler_angles_derivative_attach: np.ndarray,
    t_attach: float,
    idx: int,
    rd_approx: bool,
    rd_smoothing: bool,
    beta_approx: int = 0,
):
    """
     This function computes the ringdown approximation of the Euler angles in
     the J-frame, assuming simple precession around the final J-vector with a
     precession frequency approximately equal to the differences between the
     lowest overtone of the (2,2) and (2,1) QNM frequencies.

     Args:
        t_full (np.ndarray): Time array of the waveform modes.
        precRate (float): Precessing rate,  differences between the lowest overtone of the
                          (2,2) and (2,1) QNM frequencies. See Eq. (3.4) of 2004.09442.
        euler_angles_attach (np.ndarray): Euler angles (alpha,beta,gamma) at the attachment time.
        euler_angles_derivative_attach (np.ndarray): Time derivative of the Euler angles (alpha,beta,gamma) at the attachment time.
        t_attach (float): Attachment time of the dynamics and the ringdown
        idx (int): Index at which the attachment of the ringdown is performed
        rd_approx (bool): If True apply the approximation of the Euler angles, if False use constant angles
        rd_smoothing (bool): If True apply smoothing of the Euler angles, if False do not apply the smoothing
        beta_approx (int): If 0 use constant beta angle, otherwise use small opening angle approximation

    Returns:
        (quaternion.quaternion): Quaternion describing the ringdown approximation of the Euler angles in the J-frame
    """
    alphaAttach, betaAttach, gammaAttach = euler_angles_attach
    t_RD = t_full[idx[-1] + 1 :]

    # Approximate the Euler angles assuming simple precession
    if rd_approx:
        if rd_smoothing:

            # Smoothing: dalpha/dt and dgamma/dt are windowed with a Tanh function. The expressions for alpha and gamma are the integral of the windowed derivative
            idx_smoothing = 60 # The window is expensive so only applied on a subset of points

            t_smooth = (t_full[idx[-1] + 1 : idx[-1] + 1 + idx_smoothing] - t_attach)

            # Parameters needed for the window
            dalphaAttach, dbetaAttach, dgammaAttach = euler_angles_derivative_attach
            dalphaRD = precRate
            dgammaRD = -cos(betaAttach)*precRate
            sigma = 2.5
            ti = 2.

            # Offsets to ensure continuity
            t_offset = 0.0#t_smooth[0]
            alpha_smooth_offset = 0.5*dalphaAttach*t_offset + 0.5*dalphaRD*t_offset + (-0.125*dalphaAttach*sigma + 0.125*dalphaRD*sigma)*np.log(np.cosh((4.*(t_offset - 1.*ti))/sigma))
            gamma_smooth_offset = 0.5*dgammaAttach*t_offset + 0.5*dgammaRD*t_offset + (-0.125*dgammaAttach*sigma + 0.125*dgammaRD*sigma)*np.log(np.cosh((4.*(t_offset - 1.*ti))/sigma))

            # alpha and gamma smoothened
            alphaJ2P = alphaAttach - alpha_smooth_offset + 0.5*dalphaAttach*t_smooth + 0.5*dalphaRD*t_smooth + (-0.125*dalphaAttach*sigma + 0.125*dalphaRD*sigma)*np.log(np.cosh((4.*(t_smooth - 1.*ti))/sigma))
            gammaJ2P = gammaAttach - gamma_smooth_offset + 0.5*dgammaAttach*t_smooth + 0.5*dgammaRD*t_smooth + (-0.125*dgammaAttach*sigma + 0.125*dgammaRD*sigma)*np.log(np.cosh((4.*(t_smooth - 1.*ti))/sigma))

            # Time array for evaluating the unwindowed RD approximation
            t_RD = t_full[idx[-1] +idx_smoothing + 1 : ]

            # Ensure continuity between windowed and unwindowed regions
            t_attach_2 = t_RD[0] - t_attach
            alphaAttach2 = alphaAttach - alpha_smooth_offset + 0.5*dalphaAttach*t_attach_2 + 0.5*dalphaRD*t_attach_2 + (-0.125*dalphaAttach*sigma + 0.125*dalphaRD*sigma)*np.log(np.cosh((4.*(t_attach_2 - 1.*ti))/sigma))
            gammaAttach2 = gammaAttach - gamma_smooth_offset + 0.5*dgammaAttach*t_attach_2 + 0.5*dgammaRD*t_attach_2 + (-0.125*dgammaAttach*sigma + 0.125*dgammaRD*sigma)*np.log(np.cosh((4.*(t_attach_2 - 1.*ti))/sigma))

            #Evaluate unwindowed ringdown approximation
            t_attach_RD = t_RD[0]
            t_full += t_full[0]

            cosbetaAttach = cos(betaAttach)
            tmp = (t_RD - t_attach_RD) * precRate
            alphaJ2P = np.append(alphaJ2P,alphaAttach2 + tmp)

            betaJ2P = np.ones(len(t_smooth)+len(t_RD)) *betaAttach
            #betaJ2P = betaAttach + (t_full[idx[-1] + 1 :] - t_attach) * dbetaAttach

            gammaJ2P = np.append(gammaJ2P,gammaAttach2 - cosbetaAttach * (t_RD - t_attach_RD) * precRate)

        else:
            t_RD = t_full[idx[-1] + 1 :]
            t_full += t_full[0]

            cosbetaAttach = cos(betaAttach)
            tmp = (t_RD - t_attach) * precRate
            alphaJ2P = alphaAttach + tmp

            if beta_approx:
                y = np.exp(-tmp)
                betaJ2P = -2.0 * np.arctan2(2 * y, 1.0) + betaAttach
            else:
                betaJ2P = np.ones(len(t_RD)) * betaAttach

            gammaJ2P = gammaAttach - cosbetaAttach * (t_RD - t_attach) * precRate

        alphaJ2P = np.unwrap(alphaJ2P)
        betaJ2P = np.unwrap(betaJ2P)
        gammaJ2P = np.unwrap(gammaJ2P)

    else:

        t_RD = t_full[idx[-1] + 1 :]
        alphaJ2P = np.ones(len(t_RD)) * alphaAttach
        betaJ2P = np.ones(len(t_RD)) * betaAttach
        gammaJ2P = np.ones(len(t_RD)) * gammaAttach

    euler_angles_v1 = np.transpose([alphaJ2P, betaJ2P, gammaJ2P])
    quat_postMerger = quaternion.from_euler_angles(
        euler_angles_v1, beta=None, gamma=None
    )

    return quat_postMerger


def minimal_rotation_mine(
    R: quaternion.quaternion,
    t: quaternion.quaternion,
    v:quaternion.quaternion,
    omega: np.ndarray,
    domega: np.ndarray,
    iterations: int=2
):
    """
    Adjust frame so that there is no rotation about z' axis.
    The output of this function is a frame that rotates the z axis onto the same z' axis as the
    input frame, but with minimal rotation about that axis.  This is done by pre-composing the input
    rotation with a rotation about the z axis through an angle gamma, where
        dgamma/dt = 2*(dR/dt * z * R.conjugate()).w
    This ensures that the angular velocity has no component along the z' axis.
    Note that this condition becomes easier to impose the closer the input rotation is to a
    minimally rotating frame, which means that repeated application of this function improves its
    accuracy.  By default, this function is iterated twice, though a few more iterations may be
    called for.

    Args:
        R (quaternion.quaternion) : Quaternion describing rotation.
        t (quaternion.quaternion) : Quaternion times at which R is measured.
        v (quaternion.quaternion) : Corresponding times at which R is measured.
        omega (np.ndarray): Orbital frequency array from EOB dynamics.
        domega (np.ndarray): Time derivative of the orbital frequency array from EOB dynamics.
        iterations (int):  Repeat the minimization to refine the result. Defaults to 2.

    Returns:
        (quaternion.quaternion): Quaternion with the minimum rotation condition applied on it.

    """
    from scipy.interpolate import CubicSpline

    if iterations == 0:
        return R
    R = quaternion.as_float_array(R)

    # To compute dR/dt we first compute dR/domega
    # and then dR/dt = dR/domega*domega/dt
    # This is done to reduce the noise in the spline
    # derivative
    dRdomega = CubicSpline(omega, R).derivative()(omega)

    Rdot = dRdomega * domega[:, None]
    R = quaternion.from_float_array(R)
    Rdot = quaternion.from_float_array(Rdot)
    halfgammadot = quaternion.as_float_array(Rdot * v * np.conjugate(R))[:, 0]
    halfgamma = CubicSpline(t, halfgammadot).antiderivative()(t)
    Rgamma = np.exp(v * halfgamma)
    return minimal_rotation_mine(
        R * Rgamma, t, v, omega, domega, iterations=iterations - 1
    )


def minimal_quat(
    v_final: quaternion.quaternion,
    V: quaternion.quaternion,
    aux: quaternion.quaternion,
    t_dyn: np.ndarray,
    omega: np.ndarray
):
    """
    Compute the rotation that aligns V with v_final
    and obeys the minimal rotation condition.

    Args:
        v_final (quaternion.quaternion): z-axis quaternion of the final spin frame.
        V (quaternion.quaternion): Quaternion of LNhat in the final spin frame.
        aux (quaternion.quaternion): An optional auxiliary quaternion to help with numerical stability.
        t_dyn (np.ndarray): Time array from EOB dynamics.
        omega (np.ndarray): Orbital frequency array from EOB dynamics.

    Returns:
        (np.ndarray): Array of quaternions
    """
    omega_intrp = CubicSpline(t_dyn, omega)
    domega = omega_intrp.derivative()(t_dyn)
    if aux is not None:
        step1 = np.sqrt(-aux * v_final)
        step2 = np.sqrt(-V * aux)
        ttl = step2 * step1
    else:
        ttl = np.sqrt(-V * v_final)

    ttl_min = minimal_rotation_mine(ttl, t_dyn, v_final, omega, domega)
    return ttl_min


def inspiral_merger_quaternion_angles(
    t_dynamics: np.ndarray,
    omega_dynamics: np.ndarray,
    t_attach: float,
    Lvec_hat_attach: np.ndarray,
    Jfhat_attach: np.ndarray,
    splines: Dict[Any, Any],
    t_ref: float = None,
):
    """Wrapper function to compute the angles/quaternions necessary to perform the rotations
    from the co-precessing frame (P-frame) to the observer inertial frame (I-frame) passing
    through the final angular momentum frame (J-frame) for the inspiral part of the waveform

    Args:
        t_dynamics (np.ndarray): Time array for dynamics
        omega_dynamics (np.ndarray): EOB orbital frequency
        t_attach (float): Attachment time
        Lvec_hat_attach (np.ndarray): LN at attachment, I frame
        Jfhat_attach (np.ndarray): final spin direction, I frame
        splines (Dict[Any,Any]): Dictionary of splines
        t_ref (float, optional): Reference time if f_ref!=f_min. Defaults to None.

    Returns:
        (tuple): Quantities required to perform the rotations from the co-precessing to the inertial frame (time of
                of the dynamics, time dependent quaternion from the P-frame to the J-frame, Euler angles from the
                J-frame to the I-frame, and the quaternions during the inspiral in the J-frame)

    """

    # Apply ringdown approximation at the attachment time
    Lf_hat_v5 = Lvec_hat_attach
    Jf_hat_v5 = Jfhat_attach

    # Compute e1J, e2J, e3J triad
    e1J, e2J, e3J = SEOBBuildJframeVectors(Jf_hat_v5)
    # LN on the final dynamics time grid
    tmp_LN = splines["everything"](omega_dynamics)[:, 10:13]

    # Normalize LN_hat to ensure working with unit quaternions
    tmp_LN_norm = np.sqrt(np.einsum("ij,ij->i", tmp_LN, tmp_LN))
    tmp_LN = (tmp_LN.T / tmp_LN_norm).T

    # LN as expressed in J frame
    Ze1J = np.dot(tmp_LN, e1J)
    Ze2J = np.dot(tmp_LN, e2J)
    Ze3J = np.dot(tmp_LN, e3J)
    LN_in_J = np.c_[Ze1J, Ze2J, Ze3J]

    # Compute the Euler angles from the inertial frame to the final J-frame  (I2J angles)
    rotJ_matrix = np.vstack([e1J, e2J, e3J]).T
    quatI2J = quaternion.from_rotation_matrix(rotJ_matrix)

    # The inverse transformation, J->I
    quatJ2I = quatI2J.conjugate()
    alpha_J2I, beta_J2I, gamma_J2I = quaternion.as_euler_angles(quatJ2I)



    quatJ2P_dyn = SEOBEulerJ2PFromDynamics(t_dynamics, omega_dynamics, LN_in_J)
    # Compute Euler angles from the quaternions
    #eAngles_min = quaternion.as_euler_angles(quatJ2P_dyn).T

    #gammaJ2P_dyn = np.unwrap(eAngles_min[2])

    # We still have the residual freedom from minimal rotation condition to perform a
    # global rotation around the J_f . We fix this freedom by demanding that
    # at the reference time, the *I2P* map is the identity, i.e. the
    # inertial frame and co-precessing frame are the same. To do so,
    # we only need to ensure that at t_ref, gammaJ2P is the same as
    # gamma_J2I.
    # NB: due to quaternions being a double cover of SO(3), q and -q
    # are the same rotation.

    if t_ref is None:
        #gamma_ref = gammaJ2P_dyn[0]
        q2JP_ref = quatJ2P_dyn[0]
    else:
        #igammaJ2P_dyn = CubicSpline(t_dynamics, gammaJ2P_dyn)
        #gamma_ref = igammaJ2P_dyn(t_ref)
        q2JP_ref = quaternion.squad(quatJ2P_dyn,t_dynamics,t_ref)

    # We need to shift things so that gamma_J2P = gamma_J2I at reference time
    # The following is a rotation around final J
    #quatJ2P_dyn2 = quatJ2P_dyn*np.exp((-gamma_ref / 2 + gamma_J2I / 2) * quaternion.z)

    shift = q2JP_ref.conjugate()*quatJ2I
    quatJ2P_dyn *= shift

    eAngles_min = quaternion.as_euler_angles(quatJ2P_dyn).T
    alphaJ2P_dyn = np.unwrap(eAngles_min[0])
    betaJ2P_dyn = np.unwrap(eAngles_min[1])
    gammaJ2P_dyn = np.unwrap(eAngles_min[2])

    ialphaJ2P_dyn = CubicSpline(t_dynamics, alphaJ2P_dyn)
    ibetaJ2P_dyn = CubicSpline(t_dynamics, betaJ2P_dyn)
    igammaJ2P_dyn = CubicSpline(t_dynamics, gammaJ2P_dyn)

    alpha_attach = ialphaJ2P_dyn(t_attach)
    beta_attach = ibetaJ2P_dyn(t_attach)
    gamma_attach = igammaJ2P_dyn(t_attach)

    euler_angles_attach = [alpha_attach, beta_attach, gamma_attach]
    euler_angles_derivative_attach = [ialphaJ2P_dyn.derivative()(t_attach), ibetaJ2P_dyn.derivative()(t_attach), igammaJ2P_dyn.derivative()(t_attach)]


    # Compute sign from the angle between the final total and orbital angular momenta to check if we
    # are in the prograde or retrograde case
    cos_angle = my_dot(Jf_hat_v5, Lf_hat_v5)
    flip = 1
    if cos_angle < 0:
        flip = -1

    return t_dynamics, quatJ2P_dyn, quatI2J, euler_angles_attach,euler_angles_derivative_attach,flip


def compute_omegalm_P_frame(omegalm: complex, m: int, rotation_term: float):
    """
    Function to compute the co-precessing frame QNM frequencies
    from the J-frame QNMs as described in arXiv:2301.06558.

    Args:
        omegalm (complex): QNM frequency in the J-frame
        m (int): m index of the (l,m) waveform multipole
        rotation_term (float): (1-abs(cos[beta]))*precRate, where the precRate is the difference between the lowest overtone
                                of the 22 and 21 QNM frequencies

    Returns:
        (complex): QNM frequency in the P-frame
    """

    omega_complex_P_frame = omegalm - m * rotation_term

    return omega_complex_P_frame


@jit(nopython=True, cache=True)
def custom_swsh(beta: np.ndarray, gamma: np.ndarray, lmax: int):
    """
    Function to compute the spin-weighted spherical harmonics necessary to
    compute the polarizations in the inertial frame from the co-precessing
    frame modes [(2,|2|),(2,|1|),(3,|3|),(3,|2|),(4,|4|),(4,|3|),(5,|5|)].

    Args:
        beta (np.ndarray): Euler angle beta between the co-precessing frame and the inertial frame
        gamma (np.ndarray): Euler angle gamma between the co-precessing frame and the inertial frame
        lmax (int): Maximum ell to use

    Returns:
        (dict): Dictionary containing specific values of the spin-weighted spherical harmonics
    """

    cBH = np.cos(beta / 2.0)
    sBH = np.sin(beta / 2.0)

    cBH2 = cBH * cBH
    cBH3 = cBH2 * cBH
    cBH4 = cBH3 * cBH

    sBH2 = sBH * sBH
    sBH3 = sBH2 * sBH
    sBH4 = sBH3 * sBH

    expGamma = np.exp(1j * gamma)
    expGamma2 = expGamma * expGamma

    swsh = {}

    swsh[2, 2] = 0.5 * np.sqrt(5.0 / np.pi) * cBH4 * expGamma2
    swsh[2, 1] = np.sqrt(5.0 / np.pi) * cBH3 * sBH * expGamma
    swsh[2, -2] = 0.5 * np.sqrt(5.0 / np.pi) * sBH4 / expGamma2
    swsh[2, -1] = np.sqrt(5.0 / np.pi) * sBH3 * cBH / expGamma

    if lmax >= 3:
        cBH5 = cBH4 * cBH
        sBH5 = sBH4 * sBH
        expGamma3 = expGamma2 * expGamma
        swsh[3, 3] = -np.sqrt(10.5 / np.pi) * cBH5 * sBH * expGamma3
        swsh[3, 2] = (
            0.25 * np.sqrt(7.0 / np.pi) * cBH4 * (6.0 * (cBH2 - sBH2) - 4.0) * expGamma2
        )
        swsh[3, -3] = np.sqrt(10.5 / np.pi) * sBH5 * cBH / expGamma3
        swsh[3, -2] = (
            0.25 * np.sqrt(7.0 / np.pi) * sBH4 * (6.0 * (cBH2 - sBH2) + 4.0) / expGamma2
        )

    if lmax >= 4:
        cBH6 = cBH5 * cBH
        sBH6 = sBH5 * sBH
        expGamma4 = expGamma3 * expGamma
        swsh[4, 4] = 3.0 * np.sqrt(7.0 / np.pi) * cBH6 * sBH2 * expGamma4
        swsh[4, 3] = (
            -0.75
            * np.sqrt(3.5 / np.pi)
            * cBH5
            * sBH
            * (8.0 * (cBH2 - sBH2) - 4.0)
            * expGamma3
        )
        swsh[4, -4] = 3.0 * np.sqrt(7.0 / np.pi) * sBH6 * cBH2 / expGamma4
        swsh[4, -3] = (
            0.75
            * np.sqrt(3.5 / np.pi)
            * sBH5
            * cBH
            * (8.0 * (cBH2 - sBH2) + 4.0)
            / expGamma3
        )

    if lmax == 5:
        cBH7 = cBH6 * cBH
        sBH7 = sBH6 * sBH
        expGamma5 = expGamma4 * expGamma
        swsh[5, 5] = -np.sqrt(330.0 / np.pi) * cBH7 * sBH3 * expGamma5
        swsh[5, -5] = np.sqrt(330.0 / np.pi) * sBH7 * cBH3 / expGamma5

    return swsh

def interpolate_quats(quat: quaternion.quaternion, t_intrp: np.ndarray, t_full: np.ndarray):
    """
    Function to interpolate the quaternions from the co-precessing frame to the final
    J-frame into the equal-spacing and finer time grid of the waveform modes

    Args:
        quat (quaternion): Quaternion from the co-precessing frame to the final J-frame
        t_intrp (np.ndarray): Time array of the EOB dynamics
        t_full (np.ndarray): Time array of waveform modes


    Returns:
        (quaternion): Quaternion interpolated to the finer time grid of the waveform modes
    """
    angles = quaternion.as_euler_angles(quat)
    sp = CubicSpline(t_intrp, np.unwrap(angles.T).T)
    intrp_angles = sp(t_full)
    return quaternion.from_euler_angles(intrp_angles)



# cython: language_level=3, boundscheck=False, cdivision=True, profile=True, linetrace=True, initializedcheck=False
"""
Contains the actual RHS of the EOM, wrapped in cython. 
This allows some of the cython functions used in the RHS to be called more efficiently.
"""

cimport cython
from pyseobnr.eob.utils.containers cimport EOBParams,FluxParams
from pyseobnr.eob.waveform.waveform cimport RR_force,  RadiationReactionForce
from pyseobnr.eob.hamiltonian.Hamiltonian_v5PHM_C cimport Hamiltonian_v5PHM_C
import numpy as np
cimport numpy as np
from pyseobnr.eob.fits.fits_Hamiltonian import dSO as dSO_poly_fit


DTYPE = np.float64

# "ctypedef" assigns a corresponding compile-time type to DTYPE_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
ctypedef np.float64_t DTYPE_t

cdef extern from "numpy/npy_math.h":
    bint npy_isnan(double x)

cpdef get_rhs_prec(double t, double[::1] z, Hamiltonian_v5PHM_C H, RadiationReactionForce RR,
    double m_1, double m_2, EOBParams params):
    """
    Compute the RHS of the EOB evolution equations.
    In particular this function returns
    :math:`\\dot{r}`, :math:`\\dot{\\phi}`, :math:`\\dot{p}_{r}`, :math:`\\dot{p}_{\\phi}`.
    """
    cdef double[::1] q = z[:2]
    cdef double[::1] p = z[2:]

    cdef double chi1_LN = params.p_params.chi_1
    cdef double chi2_LN = params.p_params.chi_2

    cdef double chi1_L = params.p_params.chi1_L
    cdef double chi2_L = params.p_params.chi2_L

    cdef double H_val,omega,omega_circ,xi
    cdef double dynamics[6]
    dynamics[:] = H.dynamics(q,
        p,
        params.p_params.chi1_v,
        params.p_params.chi2_v,
        m_1,
        m_2,
        chi1_LN,
        chi2_LN,
        chi1_L,
        chi2_L,)
    H_val = dynamics[4]
    omega = dynamics[3]
    params.dynamics.p_circ[1] = p[1]

    omega_circ = H.omega(q,
        params.dynamics.p_circ,
        params.p_params.chi1_v,
        params.p_params.chi2_v,
        m_1,
        m_2,
        chi1_LN,
        chi2_LN,
        chi1_L,
        chi2_L,)

    xi = dynamics[5]

    params.p_params.omega_circ = omega_circ
    params.p_params.omega = omega
    params.p_params.H_val = H_val



    cdef (double, double) RR_f = RR.RR(q, p, omega, omega_circ, H_val, params)

    cdef double deriv[4]
    deriv[:] = [xi * dynamics[2], dynamics[3], -dynamics[0] * xi + RR_f[0], -dynamics[1] + RR_f[1]]
    return deriv



import lal
import numpy as np
import pytest

from pyseobnr.generate_waveform import GenerateWaveform


@pytest.fixture
def basic_settings():
    m1 = 50.0
    m2 = 30.0
    params_dict = {
        "mass1": m1,
        "mass2": m2,
        "spin1x": 0,
        "spin1y": 0,
        "spin1z": 0.5,
        "spin2x": 0,
        "spin2y": 0,
        "spin2z": 0.1,
        "deltaT": 1 / 2048.0,
        "deltaF": 0.125,
        "f22_start": 0.0157 / ((m1 + m2) * np.pi * lal.MTSUN_SI),
        "phi_ref": 0.0,
        "distance": 1.0,
        "inclination": np.pi / 3.0,
        "f_max": 1024.0,
        "approximant": "SEOBNRv5HM",
        "postadiabatic": False,
    }
    return params_dict


def test_mode_arrays_settings(basic_settings):
    """Checks the behaviour wrt. mode_array setting"""

    # cannot accept both
    value_error = (
        r"Only one setting can be specified between .?ModeArray.? and .?mode_array.?."
    )

    with pytest.raises(
        ValueError,
        match=value_error,
    ):
        _ = GenerateWaveform(basic_settings | {"mode_array": [], "ModeArray": []})

    with pytest.raises(
        ValueError,
        match=value_error,
    ):
        _ = GenerateWaveform(
            basic_settings
            | {"approximant": "SEOBNRv5PHM", "mode_array": [], "ModeArray": []}
        )

    calls_to_check = (
        "generate_td_modes",
        "generate_td_polarizations",
        "generate_fd_polarizations",
    )

    # works with list and tuples
    for approximant in "SEOBNRv5HM", "SEOBNRv5PHM":
        wfm_gen = GenerateWaveform(
            basic_settings
            | {"approximant": approximant, "mode_array": [(2, 2), (3, 3)]}
        )

        for func in calls_to_check:
            _ = getattr(wfm_gen, func)()

        wfm_gen = GenerateWaveform(
            basic_settings
            | {"approximant": approximant, "mode_array": ((2, 2), (3, 3))}
        )

        for func in calls_to_check:
            _ = getattr(wfm_gen, func)()

        # incorrect mode array yields an error
        wfm_gen = GenerateWaveform(
            basic_settings | {"approximant": approximant, "mode_array": [2, 2, 3, 3]}
        )
        for func in calls_to_check:
            with pytest.raises(TypeError):
                _ = getattr(wfm_gen, func)()



# cython: language_level=3
from pyseobnr.eob.utils.containers cimport EOBParams, FluxParams
cdef extern from "eob_parameters.h":
    const int PN_limit
    const int ell_max

cpdef (double, double) RR_force(double[::1] q, double[::1] p, double omega, double omega_circ,
    double H, EOBParams eob_pars)

cdef class RadiationReactionForce:
    cpdef (double, double) RR(self, double[::1] q, double[::1] p, double omega, 
        double omega_circ, double H, EOBParams eob_pars)

cdef class SEOBNRv5RRForce(RadiationReactionForce):
    cpdef (double,double) RR(self, double[::1] q,double[::1] p,double omega,double omega_circ,double H,EOBParams eob_par)



import argparse
import importlib
import os
import sys

import matplotlib.pyplot as plt
import numpy as np
from bilby.core.utils import check_directory_exists_and_if_not_mkdir
from pathos.multiprocessing import ProcessingPool as Pool
from scipy.interpolate import CubicSpline

from pyseobnr.auxiliary.sanity_checks.metrics import *


from pyseobnr.auxiliary.sanity_checks.parameters import parameters_random, parameters_random_fast
from pyseobnr.auxiliary.sanity_checks.single_waveform_tests import monotonic_quantity_test
from pyseobnr.generate_waveform import generate_modes_opt


def h22_monotonicity(
    q: float,
    chi1: float,
    chi2: float,
    plots: bool = True,
    omega0: float = 0.015,
    threshold: float = 1.0e-8,
):

    """
    Generate a SEOBNRv5 waveform with parameters (q, chi1, chi2)
    and run monotonicity check for the (2,2) mode amplitude

    Return (0,0,0) if the test passed, (q, chi1, chi2) otherwise
    """

    try:

        _, _, model = generate_modes_opt(q, chi1, chi2, omega0,debug=True)
        # Monotonicity of the (2,2) mode amplitude until merger
        h22 = np.abs(model.waveform_modes["2,2"])
        t_h = model.t
        t_merger = t_h[np.argmax(h22)]

        try:
            monotonic_quantity_test(t_h, h22, t_merger, threshold)
            return (0, 0, 0)
        except AssertionError:
            print(
                f"|h22| non monotonic for the following parameters: q = {q}, chi1 = {chi1}, chi2 = {chi2}"
            )
            if plots == True:
                label = f"q{np.round(q,2)}_s{np.round(chi1,3)}_s{np.round(chi2,3)}"
                plt_dir = "./plots_checks_h22"
                check_directory_exists_and_if_not_mkdir(plt_dir)
                plt.plot(t_h, h22)
                plt.axvline(t_merger, ls="--")
                plt.xlim(t_merger - 30, t_merger + 30)
                plt.xlabel("$t/M$")
                plt.ylabel("$|h_{22}|$")
                plt.title(f"{label}")
                plt.savefig(f"{plt_dir}/check_h22_{label}.png", dpi=150)
                plt.close()
            return (q, chi1, chi2)
    except IndexError:
        print(
            f"Error for the following parameters: q = {q}, chi1 = {chi1}, chi2 = {chi2}"
        )


def omega22_monotonicity(
    q: float,
    chi1: float,
    chi2: float,
    plots: bool = True,
    omega0: float = 0.015,
    threshold: float = 1.0e-8,
):

    """
    Generate a SEOBNRv5 waveform with parameters (q, chi1, chi2)
    and run monotonicity check for the (2,2) mode frequency

    Return (0,0,0) if the test passed, (q, chi1, chi2) otherwise
    """

    try:

        _, _, model = generate_modes_opt(q, chi1, chi2, omega0,debug=True)

        # Monotonicity of the (2,2) mode frequency until merger
        phi22 = -np.unwrap(np.angle(model.waveform_modes["2,2"]))
        t_h = model.t
        intphi22 = CubicSpline(t_h, phi22)
        omega22 = intphi22.derivative()(t_h)

        h22 = np.abs(model.waveform_modes["2,2"])
        t_merger = t_h[np.argmax(h22)]

        try:
            monotonic_quantity_test(t_h, omega22, t_merger, threshold)
            return (0, 0, 0)
        except AssertionError:
            print(
                f"Omega22 non monotonic for the following parameters: q = {q}, chi1 = {chi1}, chi2 = {chi2}"
            )
            if plots == True:
                label = f"q{np.round(q,2)}_s{np.round(chi1,3)}_s{np.round(chi2,3)}"
                plt_dir = "./plots_checks_omega22"
                check_directory_exists_and_if_not_mkdir(plt_dir)
                plt.plot(t_h, omega22)
                plt.axvline(t_merger, ls="--")
                plt.xlim(t_merger - 30, t_merger + 30)
                plt.xlabel("$t/M$")
                plt.ylabel("$\Omega_{22}$")
                plt.title(f"{label}")
                plt.savefig(f"{plt_dir}/check_om22_{label}.png", dpi=150)
                plt.close()
            return (q, chi1, chi2)
    except IndexError:
        print(
            f"Error for the following parameters: q = {q}, chi1 = {chi1}, chi2 = {chi2}"
        )


def process_one_case_h(input):
    q, chi1, chi2, plots = input
    (
        q_bad_h,
        chi1_bad_h,
        chi2_bad_h,
    ) = h22_monotonicity(q, chi1, chi2, plots)
    return np.array([q_bad_h, chi1_bad_h, chi2_bad_h])


def process_one_case_om(input):
    q, chi1, chi2, plots = input
    (
        q_bad_h,
        chi1_bad_h,
        chi2_bad_h,
    ) = omega22_monotonicity(q, chi1, chi2, plots)
    return np.array([q_bad_h, chi1_bad_h, chi2_bad_h])


if __name__ == "__main__":

    p = argparse.ArgumentParser(description="Check monotonicity across parameter space")
    p.add_argument("--points", type=int, help="Number of points", default=10000)
    p.add_argument("--q-min", type=float, help="Lower limit on q", default="1.0")
    p.add_argument("--q-max", type=float, help="Upper limit on q", default="100.0")
    p.add_argument(
        "--chi-max", type=float, help="Upper limit on |chi|", default="0.995"
    )
    p.add_argument(
        "--name",
        type=str,
        help="Name of the output file",
        default="monotonicity_test",
    )
    p.add_argument(
        "--quantity",
        type=str,
        help="Quanitiy of which to check monotonicity",
        default="amplitude",
        choices=["amplitude", "frequency"],
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    p.add_argument("--n-cpu", type=int, help="Number of cores to use", default=64)

    args = p.parse_args()

    qarr, chi1arr, chi2arr = parameters_random_fast(
        args.points,
        args.q_min,
        args.q_max,
        -args.chi_max,
        args.chi_max,
        -args.chi_max,
        args.chi_max,
    )

    lst = [(a, b, c, args.plots) for a, b, c in zip(qarr, chi1arr, chi2arr)]
    pool = Pool(args.n_cpu)

    if args.quantity == "amplitude":
        all_means = pool.map(process_one_case_h, lst)
    if args.quantity == "frequency":
        all_means = pool.map(process_one_case_om, lst)
    # """
    all_means = np.array(all_means)

    np.savetxt(f"{args.name}_{args.quantity}.dat", all_means)

    ### Plots ###

    if args.plots:

        plt_dir = f"./plots_monotonicity"
        check_directory_exists_and_if_not_mkdir(plt_dir)

        res_path = f"{args.name}_{args.quantity}.dat"
        res = np.loadtxt(res_path)

        q = res[:, 0]
        idx = np.where(q > 0.0)[0]

        q = q[idx]
        chi1 = res[:, 1][idx]
        chi2 = res[:, 2][idx]

        nu = q / (1 + q) ** 2
        m1 = q / (1 + q)
        m2 = 1 / (1 + q)
        ap = m1 * chi1 + m2 * chi2
        am = m1 * chi1 - m2 * chi2

        if len(q) > 0:
            # Cases where |h22| is not monotonic across parameter space - (q, chieff)
            plt.scatter(q, ap)
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")
            if args.quantity == "amplitude":
                plt.title("Cases where $|h_{22}|$ is not monotonic")
                plt.savefig(
                    f"{plt_dir}/monotonicity_h22.png", bbox_inches="tight", dpi=300
                )
            if args.quantity == "frequency":
                plt.title("Cases where $\Omega_{22}$ is not monotonic")
                plt.savefig(
                    f"{plt_dir}/monotonicity_om22.png", bbox_inches="tight", dpi=300
                )
            plt.close()
        else:
            print("Test passed for all cases")



"""
Contains the PN precession equations and their integration, as well as some auxiliary functions
"""

from math import log
from typing import Callable

import numpy as np

from numba import *
from numba import jit, types
from pyseobnr.auxiliary.interpolate.vector_spline import VectorSpline
from scipy.integrate import solve_ivp
from scipy.interpolate import CubicSpline

from ..hamiltonian import Hamiltonian
from .initial_conditions_aligned_precessing import computeIC_augm
from ..utils.math_ops_opt import *

# Test cythonization of PN equations
from .rhs_precessing import get_rhs_prec


@jit(
    nopython=True,
    cache=True,
)
def prec_eqns_20102022(t, z, nu, m_1, m_2, X1, X2):
    """Post-Newtonian precession equations, as
    well as the evolution of the orbital frequency

    Args:
        t (float): Time
        z (np.array): Vector of unknowns
        m_1 (float): Mass of primary
        m_2 (float): Mass of secondary
        X1 (float): m_1/(m1+m_2)
        X2 (float): m_2/(m1+m_2)

    Returns:
        np.array: RHS of equations
    """
    # print(z)
    Lh = z[:3]
    S1 = z[3:6]
    S2 = z[6:9]
    omega = z[-1]

    Lh_norm = my_norm(Lh)
    Lh /= Lh_norm

    v = omega ** (1.0 / 3)
    v2 = v * v
    v3 = v2 * v
    v4 = v3 * v
    v5 = v4 * v
    v6 = v5 * v
    v7 = v6 * v
    v8 = v7 * v
    v9 = v8 * v
    v10 = v9 * v
    v11 = v10 * v
    logv = log(v)

    dm = m_1 - m_2

    q = m_1 / m_2
    nu2 = nu * nu
    nu3 = nu2 * nu
    pi2 = np.pi * np.pi
    log2 = log(2)

    # Precession frequencies of the spins
    lNxS1 = my_cross(Lh, S1)
    lNxS2 = my_cross(Lh, S2)
    S1xS2 = my_cross(S1, S2)
    lNS1xS2 = my_dot(Lh, S1xS2)
    lNS1 = my_dot(S1, Lh)
    lNS2 = my_dot(S2, Lh)
    lNS1_2 = lNS1 * lNS1
    lNS2_2 = lNS2 * lNS2

    S1dot_v9 = (
        lNxS1 * (0.75 - (3 * dm) / 4.0 + nu / 2.0) * v5
        + (lNxS1 * ((-3 * lNS2) / 2.0 - (3 * lNS1) / (2.0 * q)) - S1xS2 / 2.0) * v6
        + lNxS1
        * (0.5625 + dm * (-0.5625 + (5 * nu) / 8.0) + (5 * nu) / 4.0 - nu2 / 24.0)
        * v7
        + lNxS1
        * (
            0.84375
            + (3 * nu) / 16.0
            + dm * (-0.84375 + (39 * nu) / 8.0 - (5 * nu2) / 32.0)
            - (105 * nu2) / 32.0
            - nu3 / 48.0
        )
        * v9
        + v8
        * (
            (nu * S1xS2) / 4.0
            + lNxS1
            * (
                lNS2 * (-0.5 + nu / 12.0)
                + lNS1 * ((-17 * nu) / 12.0 + 9 / (4.0 * q) - (15 * X2) / 4.0)
            )
        )
        + v10
        * (
            -((0.375 + (49 * nu) / 16.0 + nu2 / 48.0) * S1xS2)
            + lNxS1
            * (
                lNS2 * (-2.25 + (139 * nu) / 48.0 + (103 * nu2) / 144.0)
                + lNS1
                * (
                    (-91 * nu) / 16.0
                    + (121 * nu2) / 144.0
                    + 27 / (16.0 * q)
                    + (-6.0625 + (385 * nu) / 48.0) * X2
                )
            )
        )
    )

    S2dot_v9 = (
        lNxS2 * (0.75 + (3 * dm) / 4.0 + nu / 2.0) * v5
        + (lNxS2 * ((-3 * lNS1) / 2.0 - (3 * lNS2 * q) / 2.0) + S1xS2 / 2.0) * v6
        + lNxS2
        * (0.5625 + dm * (0.5625 - (5 * nu) / 8.0) + (5 * nu) / 4.0 - nu2 / 24.0)
        * v7
        + lNxS2
        * (
            0.84375
            + (3 * nu) / 16.0
            + dm * (0.84375 - (39 * nu) / 8.0 + (5 * nu2) / 32.0)
            - (105 * nu2) / 32.0
            - nu3 / 48.0
        )
        * v9
        + v8
        * (
            -0.25 * (nu * S1xS2)
            + lNxS2
            * (
                lNS1 * (-0.5 + nu / 12.0)
                + lNS2 * ((-17 * nu) / 12.0 + (9 * q) / 4.0 - (15 * X1) / 4.0)
            )
        )
        + v10
        * (
            (0.375 + (49 * nu) / 16.0 + nu2 / 48.0) * S1xS2
            + lNxS2
            * (
                lNS1 * (-2.25 + (139 * nu) / 48.0 + (103 * nu2) / 144.0)
                + lNS2
                * (
                    (-91 * nu) / 16.0
                    + (121 * nu2) / 144.0
                    + (27 * q) / 16.0
                    + (-6.0625 + (385 * nu) / 48.0) * X1
                )
            )
        )
    )

    # Expression PN-expanding the numerator of LNhatdot
    LNhatdot_v10 = (
        (
            (
                lNxS2
                * (
                    81
                    + 27 * nu
                    + 423 * nu2
                    - 9 * dm * (-9 - 21 * nu + 7 * nu2)
                    - 2 * nu3
                )
            )
            / (96.0 * nu)
            + (
                lNxS1
                * (
                    81
                    + 27 * nu
                    + 423 * nu2
                    + 9 * dm * (-9 - 21 * nu + 7 * nu2)
                    - 2 * nu3
                )
            )
            / (96.0 * nu)
        )
        * v10
        + (
            (dm * Lh * lNS1xS2 * (-27 + 89 * nu)) / (96.0 * nu)
            + lNxS2
            * (
                (lNS1 * (207 + 1413 * nu + 3 * dm * (9 + 71 * nu) - 32 * nu2))
                / (192.0 * nu)
                + (
                    lNS2
                    * (
                        279
                        - 366 * nu
                        + dm * (279 + 192 * nu - 439 * nu2)
                        - 1381 * nu2
                        - 138 * nu3
                    )
                )
                / (96.0 * nu2)
            )
            + lNxS1
            * (
                (lNS2 * (207 + 1413 * nu - 3 * dm * (9 + 71 * nu) - 32 * nu2))
                / (192.0 * nu)
                + (
                    lNS1
                    * (
                        279
                        - 366 * nu
                        - 1381 * nu2
                        + dm * (-279 - 192 * nu + 439 * nu2)
                        - 138 * nu3
                    )
                )
                / (96.0 * nu2)
            )
            - (9 * dm * (1 + 2 * nu) * S1xS2) / (32.0 * nu)
        )
        * v11
        + (
            -0.25 * (lNxS1 * (3 - 3 * dm + 2 * nu)) / nu
            - (lNxS2 * (3 + 3 * dm + 2 * nu)) / (4.0 * nu)
        )
        * v6
        + (
            lNxS2
            * ((3 * lNS2 * (1 + dm - 2 * nu)) / (4.0 * nu2) + (3 * lNS1) / (2.0 * nu))
            + lNxS1
            * ((3 * lNS2) / (2.0 * nu) - (3 * lNS1 * (-1 + dm + 2 * nu)) / (4.0 * nu2))
        )
        * v7
        + (
            (lNxS1 * (9 - 9 * nu - 9 * dm * (1 + nu) + 2 * nu2)) / (8.0 * nu)
            + (lNxS2 * (9 - 9 * nu + 9 * dm * (1 + nu) + 2 * nu2)) / (8.0 * nu)
        )
        * v8
        + (
            (-5 * dm * Lh * lNS1xS2) / (8.0 * nu)
            + lNxS1
            * (
                -0.0625 * (lNS2 * (99 - 15 * dm + 20 * nu)) / nu
                + (lNS1 * (-9 - 3 * dm * (-3 + nu) + 21 * nu + 4 * nu2)) / (2.0 * nu2)
            )
            + lNxS2
            * (
                -0.0625 * (lNS1 * (99 + 15 * dm + 20 * nu)) / nu
                + (lNS2 * (-9 + 3 * dm * (-3 + nu) + 21 * nu + 4 * nu2)) / (2.0 * nu2)
            )
            + (3 * dm * S1xS2) / (8.0 * nu)
        )
        * v9
    )

    # Precession frequencies of the spins
    S1sq = my_dot(S1, S1)
    S2sq = my_dot(S2, S2)
    S1S2 = my_dot(S1, S2)

    flagtailSSvdot, flagNLOSSvdot, flag4PNvdot = 1.0, 1.0, 1.0

    omega_dot = (
        -(
            -96
            * nu
            * v11
            * (
                1
                + (-2.2113095238095237 - (11 * nu) / 4.0) * v2
                + v4
                * (
                    (721 * lNS1 * lNS2) / (48.0 * nu)
                    + (34103 + 122949 * nu + 59472 * nu2) / 18144.0
                    - (247 * S1S2) / (48.0 * nu)
                    + (719 * lNS2_2 * (-nu + X1)) / (96.0 * nu2)
                    - (233 * S2sq * (-nu + X1)) / (96.0 * nu2)
                    + (719 * lNS1_2 * (-nu + X2)) / (96.0 * nu2)
                    - (233 * S1sq * (-nu + X2)) / (96.0 * nu2)
                )
                + v5
                * (
                    -0.001488095238095238 * ((4159 + 15876 * nu) * np.pi)
                    + lNS2
                    * (
                        -21.439484126984127
                        + (79 * nu) / 6.0
                        + (35.125 - 809 / (84.0 * nu)) * X1
                    )
                    + lNS1
                    * (
                        -21.439484126984127
                        + (79 * nu) / 6.0
                        + (35.125 - 809 / (84.0 * nu)) * X2
                    )
                )
                + v3
                * (
                    4 * np.pi
                    + lNS2 * (-3.1666666666666665 - (25 * X1) / (4.0 * nu))
                    + lNS1 * (-3.1666666666666665 - (25 * X2) / (4.0 * nu))
                )
                + flag4PNvdot
                * v8
                * (
                    (13 * lNS1 * lNS2 * (1615505 - 780799 * nu + 17857 * nu2))
                    / (24192.0 * nu)
                    - ((9355721 + 15851457 * nu + 3413361 * nu2) * S1S2)
                    / (72576.0 * nu)
                    + (
                        lNS2
                        * np.pi
                        * (
                            2 * nu * (-93914 + 102909 * nu)
                            + 9 * (-13320 + 50483 * nu) * X1
                        )
                    )
                    / (2016.0 * nu)
                    + (
                        lNS2_2
                        * (
                            nu * (-11888267 + 14283281 * nu - 6086003 * nu2)
                            + (11888267 + 12189903 * nu - 16745701 * nu2) * X1
                        )
                    )
                    / (48384.0 * nu2)
                    + (
                        S2sq
                        * (
                            nu * (8207303 + 19211067 * nu + 2904783 * nu2)
                            + (-8207303 - 21933531 * nu + 5264553 * nu2) * X1
                        )
                    )
                    / (145152.0 * nu2)
                    + (
                        lNS1
                        * np.pi
                        * (
                            2 * nu * (-93914 + 102909 * nu)
                            + 9 * (-13320 + 50483 * nu) * X2
                        )
                    )
                    / (2016.0 * nu)
                    + (
                        lNS1_2
                        * (
                            nu * (-11888267 + 14283281 * nu - 6086003 * nu2)
                            + (11888267 + 12189903 * nu - 16745701 * nu2) * X2
                        )
                    )
                    / (48384.0 * nu2)
                    + (
                        S1sq
                        * (
                            nu * (8207303 + 19211067 * nu + 2904783 * nu2)
                            + (-8207303 - 21933531 * nu + 5264553 * nu2) * X2
                        )
                    )
                    / (145152.0 * nu2)
                )
                + v7
                * (
                    (5 * (-2649 + 143470 * nu + 146392 * nu2) * np.pi) / 12096.0
                    + (
                        lNS2
                        * (
                            nu * (-1932041 + 2538207 * nu - 454398 * nu2)
                            + (-1195759 + 4626414 * nu - 1646001 * nu2) * X1
                        )
                    )
                    / (18144.0 * nu)
                    + (
                        lNS1
                        * (
                            nu * (-1932041 + 2538207 * nu - 454398 * nu2)
                            + (-1195759 + 4626414 * nu - 1646001 * nu2) * X2
                        )
                    )
                    / (18144.0 * nu)
                    + flagtailSSvdot
                    * (
                        (207 * lNS1 * lNS2 * np.pi) / (4.0 * nu)
                        - (12 * np.pi * S1S2) / nu
                        + (209 * lNS2_2 * np.pi * (-nu + X1)) / (8.0 * nu2)
                        - (6 * np.pi * S2sq * (-nu + X1)) / nu2
                        + (209 * lNS1_2 * np.pi * (-nu + X2)) / (8.0 * nu2)
                        - (6 * np.pi * S1sq * (-nu + X2)) / nu2
                    )
                )
                + v6
                * (
                    117.72574285227559
                    - (1712 * np.euler_gamma) / 105.0
                    - (3424 * log2) / 105.0
                    - (1712 * logv) / 105.0
                    - (56198689 * nu) / 217728.0
                    + (541 * nu2) / 896.0
                    - (5605 * nu3) / 2592.0
                    + (16 * pi2) / 3.0
                    + (451 * nu * pi2) / 48.0
                    - (lNS2 * np.pi * (74 * nu + 151 * X1)) / (6.0 * nu)
                    - (lNS1 * np.pi * (74 * nu + 151 * X2)) / (6.0 * nu)
                    + flagNLOSSvdot
                    * (
                        lNS1 * lNS2 * (-40.89930555555556 + 14433 / (224.0 * nu))
                        + (22.12847222222222 + 16255 / (672.0 * nu)) * S1S2
                        + (
                            S2sq
                            * (-(nu * (76527 + 42077 * nu)) + (76527 - 8239 * nu) * X1)
                        )
                        / (4032.0 * nu2)
                        + (
                            lNS2_2
                            * (nu * (19665 + 177611 * nu) + (-19665 + 135961 * nu) * X1)
                        )
                        / (4032.0 * nu2)
                        + (
                            S1sq
                            * (-(nu * (76527 + 42077 * nu)) + (76527 - 8239 * nu) * X2)
                        )
                        / (4032.0 * nu2)
                        + (
                            lNS1_2
                            * (nu * (19665 + 177611 * nu) + (-19665 + 135961 * nu) * X2)
                        )
                        / (4032.0 * nu2)
                    )
                )
            )
        )
        / 5.0
    )

    derivs = [
        LNhatdot_v10[0],
        LNhatdot_v10[1],
        LNhatdot_v10[2],
        S1dot_v9[0],
        S1dot_v9[1],
        S1dot_v9[2],
        S2dot_v9[0],
        S2dot_v9[1],
        S2dot_v9[2],
        omega_dot,
    ]

    return derivs


####################################################################################

def compute_omega_orb(
    t: float,
    z: np.ndarray,
    H: Hamiltonian,
    RR: Callable,
    m_1: float,
    m_2: float,
    params,
) -> np.ndarray:
    """Return the dynamics equations for aligned-spin systems

    Args:
        t (float): The current time
        z (np.array): The dynamics variables, stored as (q,p)
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR force to use. Must have same signature as Hamiltonian
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary
        params (EOBParams): Container of additional inputs

    Returns:
        np.array: The dynamics equations, including RR
    """

    n = 2
    q = z[:n]
    p = z[n:]

    omega = H.omega(
        q,
        p,
        params.p_params.chi1_v,
        params.p_params.chi2_v,
        m_1,
        m_2,
        params.p_params.chi_1,
        params.p_params.chi_2,
        params.p_params.chi1_L,
        params.p_params.chi2_L,
    )

    return omega


def rhs_wrapper(t, z, args):
    """
    Wrapper to call the cython version of the function which evaluates the
    right hand sides for the non-precessing EOB evolution

    Args:
        t (float): The current time
        z (np.array): The dynamics variables, stored as (q,p)

    Returns:
        tuple: Solution of the rhs (drdt,dphidt,dprdt,dpphidt)
    """

    return get_rhs_prec(t, z, *args)



def compute_quasiprecessing_PNdynamics_opt(
    omega_ref: float,
    omega_start: float,
    m_1: float,
    m_2: float,
    chi_1: np.ndarray,
    chi_2: np.ndarray,
    rtol: float = 1e-10,
    atol: float = 1e-12,
    backend: str="DOP853",
):
    """
    Compute the dynamics starting from omega_start, with spins
    defined at omega_ref.

    First, PN evolution equations are integrated (including backwards in time)
    to get spin and orbital angular momentum. From that we construct splines
    either in time or orbital frequency for the PN quantities. Given the splines
    we now integrate aligned-spin EOB dynamics where at every step the projections
    of the spins onto orbital angular momentum is computed via the splines.

    Args:
        omega_ref (float): Reference frequency
        omega_start (float): Starting frequency
        m_1 (float): Mass of primary
        m_2 (float): Mass of secondary
        chi_1 (np.ndarray): Dimensionless spin of the primary
        chi_2 (np.ndarray): Dimensionless spin of the secondary
        rtol (float, optional): Relative tolerance for EOB integration. Defaults to 1e-12.
        atol (float, optional): Absolute tolerance for EOB integration. Defaults to 1e-12.
        backend (str, optional): The backend to use for ODE integration. Defaults to "DOP853".

    Returns:
        tuple: Aligned-spin EOB dynamics, PN time, PN dynamics, PN splines
    """

    # Step 1,1: Initial conditions, always defined at omega_ref
    Lhat0 = [0.0, 0.0, 1.0]
    S_1 = chi_1 * m_1**2
    S_2 = chi_2 * m_2**2

    z0 = np.array([*Lhat0, *S_1, *S_2, omega_ref])

    # FIXME: check that the initial omega_ref is not too high

    mt = m_1 + m_2
    nu = m_1 * m_2 / mt / mt
    X1 = m_1 / mt
    X2 = m_2 / mt
    delta = m_1 - m_2

    def spinTaylor_stoppingTest(t, y, *args):
        """
            Function to evaluate stopping conditions of the spin-precessing PN evolution

            Args:
                t (float): Time
                y (np.ndarray): Solution of the evolution (LNhx, LNhy, LNhz, S1x, S1y, S1z, S2x, S2y, S2z, omega)

            Returns:
                (int) If 0 stop the evolution, otherwise continue
        """

        # Read the solution
        LNhx, LNhy, LNhz, S1x, S1y, S1z, S2x, S2y, S2z, omega = y

        r = omega ** (-2.0 / 3)

        omegaEnd = 0.0
        omegaStart = omega_ref

        omegadiff = omega - spinTaylor_stoppingTest.omegas[-1]

        # Check d^2omega/dt^2 > 0 (true iff current_domega - prev_domega > 0)
        ddomega = omegadiff

        spinTaylor_stoppingTest.omegas.append(omega)

        v = (omega)**(1./3.)

        if np.isnan(omega):
            yout = 0

        elif v >= 1.0:  #  v/c >= 1!
            yout = 0

        elif ddomega <= 0.0 or (abs(ddomega) < 1e-9 and r<6.0):  #  // d^2omega/dt^2 <= 0!
            #printf(f"ddomega <= 0.0 :  omega = {omega}, spinTaylor_stoppingTest.omegas[-1] = {spinTaylor_stoppingTest.omegas[-1]}")
            yout = 0

        # Empirical bound on the PN integration (typically it ends earlier)
        elif omega > 0.35:
            yout = 0

        else:
            yout = 1

        return yout

    spinTaylor_stoppingTest.omegas = [omega_ref]
    spinTaylor_stoppingTest.omegaPeaks = 0

    spinTaylor_stoppingTest.terminal = True
    spinTaylor_stoppingTest.direction = -1

    res_forward = solve_ivp(
        prec_eqns_20102022,
        (0, 2e9),
        z0,
        events=spinTaylor_stoppingTest,
        args=(nu, m_1, m_2, X1, X2),
        rtol=rtol,
        atol=atol,
        method=backend,
    )


    if omega_start < omega_ref:
        # We want to start at a lower frequency than omega_ref
        # Thus we first integrate backwards in time
        def term_backwards(t, y, *args):
            return y[-1] - omega_start

        term_backwards.terminal = True

        res_back = solve_ivp(
            prec_eqns_20102022,
            (0, -2e9),
            z0,
            events=term_backwards,
            args=(nu, m_1, m_2, X1, X2),
            rtol=rtol,
            atol=atol,
            method=backend,
        )

        # Combine the forwards and backward pieces
        combined_t = np.concatenate((res_back.t[::-1][:-1], res_forward.t[1:]))
        combined_y = np.vstack((res_back.y.T[::-1][:-1], res_forward.y.T[1:, :]))

        # Set t=0 at the lowest frequency
        combined_t -= combined_t[0]
    else:

        combined_t = res_forward.t[:]
        combined_y = res_forward.y.T[:]

    # Use this to avoid interpolation errors due to repeating the last point
    tdiff_abs = abs(combined_t[-1] - combined_t[-2])
    if tdiff_abs < 1e-10:
        combined_t = combined_t[:-1]
        combined_y = combined_y[:-1]

    omega_diff = combined_y[-1, -1] - combined_y[-2, -1]
    if omega_diff < 0:
        combined_t = combined_t[:-1]
        combined_y = combined_y[:-1]

    return combined_t, combined_y

def build_splines_PN(
    combined_t: np.ndarray,
    combined_y: np.ndarray,
    m_1: float,
    m_2: float,
    omega_start: float,
):
    """
    Compute splines of the spin-precessing PN dynamics

    Args:
        combined_t (np.ndarray): Time array
        combined_y (np.ndarray): Solution of the spin-precessing PN evolution (LN_hat,S1,S2) vectors
        m_1 (float): Mass of primary
        m_2 (float): Mass of secondary
        omega_start (float): Starting orbital frequency

    Returns:
        (dict): Dictionary containing the splines in orbital frequency of the vector components of
                the spins, LN and L as well as the spin projections onto LN and L
    """

    # Build the splines for PN quantities
    splines = {}
    s1_vec = combined_y[:, 3:6]
    s2_vec = combined_y[:, 6:9]
    chi1_v = s1_vec / m_1**2
    chi2_v = s2_vec / m_2**2

    omega = combined_y[:, -1]
    LN_hat = combined_y[:, :3]

    # Renormalize to account for numerical error
    LN_hat_norm = np.sqrt(np.einsum("ij,ij->i", LN_hat, LN_hat))
    LN_hat = (LN_hat.T / LN_hat_norm).T

    mtot = m_1 + m_2
    nu = m_1 * m_2 / mtot / mtot
    X1 = m_1 / mtot
    X2 = m_2 / mtot

    v = omega ** (1.0 / 3.0)

    s1_lN = np.einsum("ij,ij->i", s1_vec, LN_hat)
    s2_lN = np.einsum("ij,ij->i", s2_vec, LN_hat)

    chi1_LN = np.einsum("ij,ij->i", chi1_v, LN_hat)
    chi2_LN = np.einsum("ij,ij->i", chi2_v, LN_hat)

    s1s2 = np.einsum("ij,ij->i", s1_vec, s2_vec)
    s1sq = np.einsum("ij,ij->i", s1_vec, s1_vec)
    s2sq = np.einsum("ij,ij->i", s2_vec, s2_vec)

    q = m_1 / m_2

    # Compute the orbital angular momentum from LN_hat, S1 and S2 up to 4PN
    L_vec = compute_Lvec_35PN_vec_opt(
        q, nu, X1, X2, v, s1_vec, s2_vec, LN_hat, s1sq, s2sq, s1s2, s1_lN, s2_lN
    )

    # Compute L_hat unit vector
    L_vec_norm = np.sqrt(np.einsum("ij,ij->i", L_vec, L_vec))
    L_hat = (L_vec.T / L_vec_norm).T

    # Compute the projections of the dimensionless spins onto L_hat
    chi1_L = np.einsum("ij,ij->i", chi1_v, L_hat)
    chi2_L = np.einsum("ij,ij->i", chi2_v, L_hat)

    # Create spline dictionary
    all_array = np.c_[chi1_LN, chi2_LN, chi1_L, chi2_L, chi1_v, chi2_v, LN_hat, L_vec]
    splines["everything"] = CubicSpline(omega, all_array)

    return splines


def compute_Lvec_35PN_vec_opt(
                              q:float,
                              nu:float,
                              X1:float,
                              X2:float,
                              v: np.ndarray,
                              S1: np.ndarray,
                              S2: np.ndarray,
                              Lh: np.ndarray,
                              S1sq: np.ndarray,
                              S2sq: np.ndarray,
                              S1S2: np.ndarray,
                              lNS1: np.ndarray,
                              lNS2: np.ndarray
):
    """
    Compute orbital angular momentum vector, L, up to 4PN using the Newtonian orbital angular momentum
    unit vector and the spin vectors

    Args:
        q (float): mass ratio > 1
        nu (float): symmetric mass ratio m1/m2> 1
        X1 (float): Mass of primary divided total mass m1/(m1+m2)
        X2 (float): Mass of secondary divided total mass m2/(m1+m2)
        v (float): Post-Newtonian velocity parameter, v= (orbital frequency)^(1/3)
        X2 (float): Mass of secondary
        S1 (np.ndarray): Dimensionful spin vector of the primary
        S2 (np.ndarray): Dimensionful spin vector of the secondary
        Lh (np.ndarray): Newtonian orbital angular momentum unit vector
        S1sq (np.ndarray): Dot product of the spin vector of the primary
        S2sq (np.ndarray): Dot product of the spin vector of the secondary
        S1S2 (np.ndarray): Dot product of the spin vectors of the primary and the secondary
        lNS1 (np.ndarray): Dot product of the Newtonian orbital angular momentum unit vector and the spin vectors of the primary
        lNS2 (np.ndarray): Dot product of the Newtonian orbital angular momentum unit vector and the spin vectors of the secondary

    Returns:
        (dict): Dictionary containing the splines in orbital frequency of the vector components of
                the spins, LN and L as well as the spin projections onto LN and L
    """

    # Compute necessary powers of v
    v2 = v * v
    v3 = v2 * v
    v4 = v3 * v
    v5 = v4 * v
    v6 = v5 * v
    v7 = v6 * v
    v8 = v7 * v
    logv = np.log(v)

    nu2 = nu * nu
    nu3 = nu2 * nu
    nu4 = nu3 * nu

    lNS1_2 = lNS1 * lNS1
    lNS2_2 = lNS2 * lNS2

    pi2 = np.pi * np.pi
    log2 = log(2.0)

    LhT = Lh.T
    S1T = S1.T
    S2T = S2.T

    # Non-spinning term
    absVecL0 = nu * (
        1 / v
        + ((9 + nu) * v) / 6.0
        + ((81 - 57 * nu + nu2) * v3) / 24.0
        + ((10935 + 1674 * nu2 + 7 * nu3 + 9 * nu * (-6889 + 246 * pi2)) * v5) / 1296.0
        + v7
        * (
            (
                13778100
                - 77400 * nu3
                - 1100 * nu4
                - 900 * nu2 * (-71207 + 2706 * pi2)
                - 27
                * nu
                * (-395476 + 983040 * np.euler_gamma + 96825 * pi2 + 1966080 * log2)
            )
            / 622080.0
            - (128 * nu * logv) / 3.0
        )
    )

    vecL0 = absVecL0 * LhT

    # SO term up to NNLO
    vecLSO = (
        v2
        * (
            (S2T * (-nu - 3 * X1)) / 4.0
            + (S1T * (-nu - 3 * X2)) / 4.0
            + LhT
            * ((-7 * lNS2 * (nu + 3 * X1)) / 12.0 - (7 * lNS1 * (nu + 3 * X2)) / 12.0)
        )
        + v4
        * (
            (S2T * ((-45 + nu) * nu + 3 * (-9 + 10 * nu) * X1)) / 48.0
            + (S1T * ((-45 + nu) * nu + 3 * (-9 + 10 * nu) * X2)) / 48.0
            + LhT
            * (
                (11 * lNS2 * ((-45 + nu) * nu + 3 * (-9 + 10 * nu) * X1)) / 144.0
                + (11 * lNS1 * ((-45 + nu) * nu + 3 * (-9 + 10 * nu) * X2)) / 144.0
            )
        )
        + v6
        * (
            (S2T * (nu * (-243 + 165 * nu + nu2) - 3 * (27 - 156 * nu + 5 * nu2) * X1))
            / 96.0
            + (
                S1T
                * (nu * (-243 + 165 * nu + nu2) - 3 * (27 - 156 * nu + 5 * nu2) * X2)
            )
            / 96.0
            + LhT
            * (
                (
                    5
                    * lNS2
                    * (
                        nu * (-243 + 165 * nu + nu2)
                        - 3 * (27 - 156 * nu + 5 * nu2) * X1
                    )
                )
                / 96.0
                + (
                    5
                    * lNS1
                    * (
                        nu * (-243 + 165 * nu + nu2)
                        - 3 * (27 - 156 * nu + 5 * nu2) * X2
                    )
                )
                / 96.0
            )
        )
    )

    # SS term up to NNLO
    vecLSS = (
        (
            (lNS2 / 2.0 + lNS1 / (2.0 * q)) * S1T
            + (lNS1 / 2.0 + (lNS2 * q) / 2.0) * S2T
            + LhT
            * (
                2 * lNS1 * lNS2
                + lNS1_2 / q
                + lNS2_2 * q
                - S1S2
                - S1sq / (2.0 * q)
                - (q * S2sq) / 2.0
            )
        )
        * v3
        + v5
        * (
            S2T
            * (
                lNS1 * (1.25 - (7 * nu) / 24.0)
                + (lNS2 * (5 * nu + 33 * q - 12 * X1)) / 24.0
            )
            + LhT
            * (
                lNS1 * lNS2 * (-1.1666666666666667 + (13 * nu) / 36.0)
                + (2 * nu * S1S2) / 3.0
                + S2sq * (-0.3333333333333333 * nu + q - (5 * X1) / 3.0)
                + (lNS2_2 * (121 * nu - 315 * q + 396 * X1)) / 72.0
                + S1sq * (-0.3333333333333333 * nu + 1 / q - (5 * X2) / 3.0)
                + lNS1_2 * ((121 * nu) / 72.0 - 35 / (8.0 * q) + (11 * X2) / 2.0)
            )
            + S1T
            * (
                lNS2 * (1.25 - (7 * nu) / 24.0)
                + (lNS1 * (33 + 5 * nu * q - 12 * q * X2)) / (24.0 * q)
            )
        )
        + v7
        * (
            S2T
            * (
                (lNS1 * (1080 - 3141 * nu - 446 * nu2)) / 576.0
                + (
                    lNS2
                    * ((1689 - 235 * nu) * nu + 189 * q - 3 * (224 + 339 * nu) * X1)
                )
                / 288.0
            )
            + S1T
            * (
                (lNS2 * (1080 - 3141 * nu - 446 * nu2)) / 576.0
                - (
                    lNS1
                    * (
                        -189
                        + nu * (-1689 + 235 * nu) * q
                        + 3 * (224 + 339 * nu) * q * X2
                    )
                )
                / (288.0 * q)
            )
            + LhT
            * (
                (lNS1 * lNS2 * (3240 + 1083 * nu - 722 * nu2)) / 864.0
                - (5 * (18 + 147 * nu + nu2) * S1S2) / 72.0
                + (
                    lNS2_2
                    * ((3123 - 505 * nu) * nu - 2997 * q + (10746 - 8499 * nu) * X1)
                )
                / 864.0
                + (5 * S2sq * (nu * (165 + nu) + 27 * q + (-273 + 59 * nu) * X1))
                / 144.0
                + (5 * S1sq * (27 + nu * (165 + nu) * q + (-273 + 59 * nu) * q * X2))
                / (144.0 * q)
                - (
                    lNS1_2
                    * (
                        2997
                        + nu * (-3123 + 505 * nu) * q
                        + 3 * (-3582 + 2833 * nu) * q * X2
                    )
                )
                / (864.0 * q)
            )
        )
    )

    vecL = vecL0 + vecLSO + vecLSS

    return vecL.T



import math

import numpy as np
import pandas as pd


def compare_frames(test_frame, reference_frame, known_differences, time_tolerance=5e-4):
    # sanity checks
    # all returned values are numbers
    assert test_frame.isna().any(axis=None).item() is False
    # all time steps are unique
    assert test_frame.t.is_unique and reference_frame.t.is_unique
    # time starts with 0
    assert test_frame.t[0] == 0
    # time is all positive
    assert (test_frame.t[1:] > 0).all()
    # time is ordered and strictly increasing
    assert (test_frame.t.diff()[1:] > 0).all()
    # time steps are decreasing non monotonically
    # this is not true
    # assert (frame_ehm.t.diff()[1:].diff().diff() < 0).all()

    # we generate the same number of elements in the dynamic
    assert len(reference_frame.columns) == reference_frame.shape[1]

    # dynamics stop at more or less the same time
    assert (
        test_frame[["t"]].max() - reference_frame[["t"]].max()
    ).abs().item() < time_tolerance

    # we create a new index composite of both index, keep unique values
    new_index_t = np.array(
        sorted(pd.concat((reference_frame.t, test_frame.t)).unique())
    )

    # we now inject the time steps of the reference frame into the current frame
    test_frame_missing_values = test_frame.set_index("t").reindex(new_index_t)

    # we interpolate the missing values (method is important, index takes the time
    # point for performing the interpolation but is not precise enough)
    test_frame_missing_values_interpolated = test_frame_missing_values.interpolate(
        # method="index"
        method="cubic",
        order=3,
        limit_area="inside",
    )

    # all timesteps of the reference frame have now a proper value
    if test_frame.t.max() > reference_frame.t.max():
        assert (
            test_frame_missing_values_interpolated.loc[reference_frame.t]
            .isna()
            .any(axis=None)
            .item()
            is False
        )
    else:
        # drop the last element in this case
        assert (
            test_frame_missing_values_interpolated.loc[reference_frame.t]
            .isna()
            .iloc[:-1]
            .any(axis=None)
            .item()
            is False
        )

    test_frame_projected = test_frame_missing_values_interpolated.loc[reference_frame.t]

    # dynamics do not necessarily have the same subdivision
    reference_frame_t = reference_frame.set_index("t")
    for idx_column, column in enumerate(reference_frame_t.columns):
        # transforms 5.755399e-05 to 1e-4
        abs_diff = known_differences[column]
        abs_diff = 10 ** math.ceil(math.log10(abs_diff))

        assert (
            reference_frame_t[[column]] - test_frame_projected[[column]]
        ).abs().max().item() < abs_diff, f"column {column}"
